from typing import (
    List,
    Optional,
)

from pygls.workspace import Workspace

from galaxyls.services.tools.common import TestsDiscoveryService
from galaxyls.services.tools.document import GalaxyToolXmlDocument
from galaxyls.services.validation import DocumentValidator
from galaxyls.services.xml.document import XmlDocument
from galaxyls.services.xml.parser import XmlDocumentParser
from galaxyls.types import (
    TestInfoResult,
    TestSuiteInfoResult,
)


class ToolTestsDiscoveryService(TestsDiscoveryService):
    document_validator = DocumentValidator()

    def discover_tests_in_workspace(self, workspace: Workspace) -> List[TestSuiteInfoResult]:
        rval: List[TestSuiteInfoResult] = []
        for doc_uri in workspace.documents:
            document = workspace.get_document(doc_uri)
            if self.document_validator.is_tool_document(document):
                xml_document = XmlDocumentParser().parse(document)
                test_suite = self._get_test_suite_from_document(xml_document)
                if test_suite:
                    rval.append(test_suite)
        return rval

    def discover_tests_in_document(self, xml_document: XmlDocument) -> Optional[TestSuiteInfoResult]:
        test_suite = self._get_test_suite_from_document(xml_document)
        return test_suite

    def _get_test_suite_from_document(self, xml_document: XmlDocument) -> Optional[TestSuiteInfoResult]:
        tool = GalaxyToolXmlDocument.from_xml_document(xml_document)
        tool_id = tool.get_tool_id()
        tests_range = tool.get_tests_range()
        if tool_id and tests_range:
            tests = tool.get_tests()
            test_cases: List[TestInfoResult] = []
            id = 1
            for test in tests:
                range = xml_document.get_full_range(test)
                if range:
                    test_cases.append(
                        TestInfoResult(
                            tool_id=tool_id,
                            test_id=str(id),
                            uri=xml_document.document.uri,
                            range=range,
                        ),
                    )
                    id += 1
            return TestSuiteInfoResult(
                tool_id=tool_id,
                uri=xml_document.document.uri,
                range=tests_range,
                children=test_cases,
            )
        return None



from typing import List

from lsprotocol.types import DocumentLink

from galaxyls.services.tools.document import GalaxyToolXmlDocument
from galaxyls.services.xml.document import XmlDocument
from galaxyls.services.xml.nodes import XmlElement
from galaxyls.services.xml.utils import convert_document_offsets_to_range


class DocumentLinksProvider:
    """Provides links to external resources defined in the document."""

    def get_document_links(self, xml_document: XmlDocument) -> List[DocumentLink]:
        tool = GalaxyToolXmlDocument.from_xml_document(xml_document)
        test_data_file_links = self._get_test_data_file_links(tool)
        return test_data_file_links

    def _get_test_data_file_links(self, tool: GalaxyToolXmlDocument) -> List[DocumentLink]:
        result: List[DocumentLink] = []
        for test in tool.get_tests():
            params = filter(lambda e: e.name == "param", test.elements)
            for param in params:
                value_attribute = param.attributes.get("value")
                filename = param.get_attribute_value("value")
                if not value_attribute or not value_attribute.value or not filename:
                    # Must have a value
                    continue

                if not self._is_data_input_param(param, tool):
                    continue

                test_data_file_path = tool.get_test_data_path() / filename
                start_offset, end_offset = value_attribute.value.get_unquoted_content_offsets()
                link_range = convert_document_offsets_to_range(tool.xml_document.document, start_offset, end_offset)
                result.append(
                    DocumentLink(
                        target=test_data_file_path.as_uri(),
                        range=link_range,
                    )
                )
        return result

    def _is_data_input_param(self, param: XmlElement, tool: GalaxyToolXmlDocument) -> bool:
        param_name = param.get_attribute_value("name")
        inputs = tool.get_inputs()
        for input in inputs:
            input_name = input.get_attribute_value("name")
            if input_name == param_name:
                input_type = input.get_attribute_value("type")
                return input_type == "data"
        return False



"""Utilities to validate Galaxy xml tool wrappers.
"""

from pathlib import Path
from typing import List

from lsprotocol.types import (
    Diagnostic,
    DiagnosticRelatedInformation,
    Location,
    Position,
    Range,
)
from lxml import etree

from galaxyls.constants import DiagnosticCodes
from galaxyls.services.tools.document import GalaxyToolXmlDocument
from galaxyls.services.xml.document import XmlDocument

EXPAND_DOCUMENT_URI_SUFFIX = "%20%28Expanded%29"


class GalaxyToolSchemaValidationService:
    """Service providing diagnostics for errors in the XML validation."""

    diagnostics_source = "Galaxy Schema Validator"

    def __init__(self, xsd_schema: etree.XMLSchema):
        """Initializes the validator"""
        self.xsd_schema = xsd_schema

    def validate_document(self, xml_document: XmlDocument) -> List[Diagnostic]:
        """Validates the XML document and returns a list of diagnostics
        if there are any problems.

        Args:
            xml_document (XmlDocument): The XML document. Can be a tool wrapper or macro
            definition file.

        Returns:
            List[Diagnostic]: The list of issues found in the document.
        """

        syntax_errors = self._check_syntax(xml_document)
        if syntax_errors:
            return syntax_errors

        if not xml_document.is_tool_file:
            return []
        try:
            return self._validate(xml_document)
        except ExpandMacrosFoundException:
            return self._validate_expanded(xml_document)
        except etree.XMLSyntaxError as e:
            return self._build_diagnostics_from_syntax_error(e)

    def _check_syntax(self, xml_document: XmlDocument) -> List[Diagnostic]:
        """Check if the XML document contains any syntax error and returns it in a list.

        Args:
            document (Document): The XML document.

        Returns:
            List[Diagnostic]: The list containing the syntax error found or an empty list.
        """
        try:
            etree.fromstring(xml_document.document.source)
            return []
        except etree.XMLSyntaxError as e:
            return self._build_diagnostics_from_syntax_error(e)

    def _validate_expanded(self, xml_document: XmlDocument) -> List[Diagnostic]:
        """Validates the document after loading all the macros referenced and expands them.

        Args:
            document (Document): [description]
            xml_tree (etree.ElementTree): [description]

        Returns:
            List[Diagnostic]: [description]
        """
        try:
            if xml_document.xml_tree_expanded:
                self.xsd_schema.assertValid(xml_document.xml_tree_expanded)
            return []
        except etree.DocumentInvalid as e:
            return self._build_diagnostics_for_expanded_macros(xml_document, e)

        except etree.XMLSyntaxError as e:
            return self._build_diagnostics_for_macros_file_syntax_error(xml_document, e)

        except AssertionError as e:
            return self._build_diagnostics_for_assertion_error(xml_document, e)

    def _validate(self, xml_document: XmlDocument) -> List[Diagnostic]:
        """Validates an XML document against the XSD schema.
        Returns:
            List[Diagnostic]: The diagnosis results
        """
        try:
            if xml_document.xml_tree:
                self.xsd_schema.assertValid(xml_document.xml_tree)
            return []
        except etree.DocumentInvalid as e:
            return self._build_diagnostics(e.error_log, xml_document)

    def _build_diagnostics(self, error_log: etree._ListErrorLog, xml_document: XmlDocument) -> List[Diagnostic]:
        """Gets a list of diagnostics from the XSD validation error log.

        Args:
            error_log (etree._ListErrorLog): The error log generated after XSD schema validation.
            xml_tree (Optional[etree.ElementTree], optional): The element tree associated with
            the error log. Defaults to None.

        Raises:
            ExpandMacrosFoundException: This is raised when a macro ``expand`` element is found in the error log.
            The ``expand`` element is not part of the XSD so, when this exception is raised, we know that
            the document must expand the macros before validation.

        Returns:
            List[Diagnostic]: The list of resulting diagnostic items found in the error log.
        """
        diagnostics = []
        for error in error_log.filter_from_errors():
            if "expand" in error.message:
                raise ExpandMacrosFoundException()

            range = xml_document.get_element_range_from_xpath_or_default(error.path)
            result = Diagnostic(
                range=range,
                message=error.message,
                source=self.diagnostics_source,
            )
            diagnostics.append(result)
        return diagnostics

    def _build_diagnostics_from_syntax_error(self, error) -> List[Diagnostic]:
        """Builds a Diagnostic element from a lxml syntax error.

        Args:
            error: The syntax error.

        Returns:
            Diagnostic: The converted Diagnostic item.
        """
        result = Diagnostic(
            range=Range(
                start=Position(line=error.lineno - 1, character=error.position[0] - 1),
                end=Position(line=error.lineno - 1, character=error.position[1] - 1),
            ),
            message=error.msg,
            source=self.diagnostics_source,
        )
        return [result]

    def _build_diagnostics_for_macros_file_syntax_error(self, xml_document: XmlDocument, syntax_error) -> List[Diagnostic]:
        tool = GalaxyToolXmlDocument.from_xml_document(xml_document)
        result = Diagnostic(
            range=tool.get_import_macro_file_range(syntax_error.filename) or XmlDocument.DEFAULT_RANGE,
            message=syntax_error.msg,
            source=self.diagnostics_source,
            related_information=[
                DiagnosticRelatedInformation(
                    message="Syntax error found on imported file.",
                    location=Location(
                        uri=Path(syntax_error.filename).as_uri(),
                        range=Range(
                            start=Position(line=syntax_error.lineno - 1, character=syntax_error.offset),
                            end=Position(line=syntax_error.lineno - 1, character=syntax_error.offset),
                        ),
                    ),
                )
            ],
        )
        return [result]

    def _build_diagnostics_for_expanded_macros(self, xml_document: XmlDocument, invalid_document_error) -> List[Diagnostic]:
        virtual_uri = xml_document.document.uri.replace("file", "gls-expand")
        diagnostics = []
        for error in invalid_document_error.error_log.filter_from_errors():
            related_info: List[DiagnosticRelatedInformation] = []
            elem_in_main_doc = xml_document.get_element_from_xpath(error.path)
            if elem_in_main_doc is None:
                elem_in_main_doc = xml_document.get_element_from_xpath("/tool/macros")
                related_info = [
                    DiagnosticRelatedInformation(
                        message=(
                            "The validation error ocurred on the expanded version of "
                            "the document, i.e. after replacing macros. "
                            "Click here to preview the expanded document."
                        ),
                        location=Location(
                            uri=f"{virtual_uri}{EXPAND_DOCUMENT_URI_SUFFIX}",
                            range=Range(
                                start=Position(line=error.line - 1, character=error.column),
                                end=Position(line=error.line - 1, character=error.column),
                            ),
                        ),
                    )
                ]
            result = Diagnostic(
                range=xml_document.get_internal_element_range_or_default(elem_in_main_doc),
                message=error.message,
                source=self.diagnostics_source,
                code=DiagnosticCodes.INVALID_EXPANDED_TOOL,
                related_information=related_info,
            )
            diagnostics.append(result)
        return diagnostics

    def _build_diagnostics_for_assertion_error(self, xml_document: XmlDocument, error: AssertionError) -> List[Diagnostic]:
        result = Diagnostic(
            range=xml_document.get_default_range(),
            message=str(error),
            source=self.diagnostics_source,
        )
        return [result]


class ExpandMacrosFoundException(Exception):
    """This exceptions indicates that the current tool contains
    macros and they should be expanded before validation.
    """



""" This code is based on the Eclipse/Lemminx XML language server implementation:
https://github.com/eclipse/lemminx/tree/master/org.eclipse.lemminx/src/main/java/org/eclipse/lemminx/dom

Only the minimum subset of the XML dialect used by Galaxy tool wrappers is supported.
"""

from typing import (
    Callable,
    List,
)

from lsprotocol.types import (
    Position,
    Range,
)
from pygls.workspace import Document

from .constants import (
    _LAN,
    NEW_LINE,
    WHITESPACE_CHARS,
)


def convert_document_offset_to_line(document: Document, offset: int) -> int:
    """Converts the given offset in the document to the corresponding line.

    Args:
        document (Document): The source document.
        offset (int): The character offset inside the document.

    Returns:
        int: The resulting document line.
    """
    line = max(document.source.count(NEW_LINE, 0, offset), 0)
    return line


def convert_document_offset_to_position(document: Document, offset: int) -> Position:
    """Converts the given offset in the document to a line/character based Position.

    Args:
        document (Document): The source document.
        offset (int): The character offset inside the document.

    Returns:
        Position: The resulting Position with line and character offset.
    """
    line = convert_document_offset_to_line(document, offset)
    line_offset = max(document.source.rfind(NEW_LINE, 0, offset), 0)
    character = offset - line_offset
    if line > 0:
        character -= 1
    return Position(line=line, character=character)


def convert_document_offsets_to_range(document: Document, start_offset: int, end_offset: int) -> Range:
    """Converts the given start and end offset positions in the document to a
    position Range based on line numbers.

    Args:
        start_offset (int): The start offset of the range
        end_offset (int): The end offset of the range

    Returns:
        Range: The resulting Range with the correct line number and character offset
    """
    return Range(
        start=convert_document_offset_to_position(document, start_offset),
        end=convert_document_offset_to_position(document, end_offset),
    )


class MultiLineStream:
    """Represents a multi-line stream of characters."""

    def __init__(self, source: str, position: int = 0) -> None:
        self._source = source
        self._position = position
        self._len = len(source)

    def eos(self) -> bool:
        """Indicates that the stream is at the end position."""
        return self._len <= self._position

    def get_source(self) -> str:
        """Gets the source text."""
        return self._source

    def pos(self) -> int:
        """Gets the current stream position."""
        return self._position

    def advance(self, n: int) -> None:
        """Advances the stream given number of positions."""
        self._position += n

    def go_to_end(self) -> None:
        """Sets the stream position to the end of the stream."""
        self._position = self._len

    def peek_char(self, n: int = 0) -> int:
        """Gets the value of the next position in the stream without advancing it."""
        try:
            return ord(self._source[self._position + n])
        except IndexError:
            return -1

    def advance_if_char(self, ch: int) -> bool:
        """If the next character in the stream matches the given character, the stream advances one position."""
        if ch == self.peek_char():
            self._position += 1
            return True
        return False

    def advance_if_chars(self, ch: List[int]) -> bool:
        """If the next characters in the stream matches the given sequence of characters, the stream advances
        the length of the sequence."""
        ch_length = len(ch)
        if self._position + ch_length > self._len:
            return False
        i = 0
        for i in range(ch_length):
            if self._source[self._position + i] != chr(ch[i]):
                return False
        self.advance(i)
        return True

    def advance_until_char(self, ch: int) -> bool:
        """Advances the stream until it founds a character matching the given."""
        while self._position < self._len:
            if ord(self._source[self._position]) == ch:
                return True
            self.advance(1)
        return False

    def advance_until_chars(self, ch: List[int]) -> bool:
        """Advances the stream until it founds a list of character matching the given."""
        ch_length = len(ch)
        while self._position + ch_length <= self._len:
            i = 0
            while i < ch_length and self._source[self._position + i] == chr(ch[i]):
                i += 1
            if i == ch_length:
                return True
            self.advance(i or 1)
        self.go_to_end()
        return False

    def advance_while_char_in(self, list: List[int]) -> int:
        """Advances the stream if the characters are like any of the characters in the given list."""
        pos_now = self._position
        while self._position < self._len and ord(self._source[self._position]) in list:
            self._position += 1
        return self._position - pos_now

    def advance_until_char_or_new_tag(self, ch: int) -> bool:
        """Advances the stream until it finds the given character or the '<' (new tag character)."""
        while self._position < self._len:
            if self.peek_char() in [ch, _LAN]:
                return True
            self.advance(1)
        return False

    def advance_until_chars_or_new_tag(self, ch: List[int]) -> bool:
        """Advances the stream until it finds the given sequence of characters or the '<' (new tag character)."""
        ch_length = len(ch)
        while self._position + ch_length <= self._len:
            i = 0
            if self.peek_char() == _LAN:
                return True
            while i < ch_length and self._source[self._position + i] == chr(ch[i]):
                i += 1
            if i == ch_length:
                return True
            self.advance(i or 1)
        self.go_to_end()
        return False

    def advance_while_char(self, predicate: Callable[[str], bool]) -> int:
        """Advances the stream while the given condition is True."""
        pos_now = self._position
        while self._position < self._len and predicate(self._source[self._position]):
            self._position += 1
        return self._position - pos_now

    def skip_whitespace(self) -> bool:
        """Advances the stream while any kind of white space character is found."""
        n = self.advance_while_char_in(WHITESPACE_CHARS)
        return n > 0



from typing import (
    List,
    Optional,
)

from lsprotocol.types import (
    DocumentSymbol,
    SymbolKind,
)

from galaxyls.services.xml.document import XmlDocument
from galaxyls.services.xml.nodes import (
    XmlAttribute,
    XmlElement,
    XmlSyntaxNode,
)
from galaxyls.services.xml.utils import convert_document_offsets_to_range


class DocumentSymbolsProvider:
    """Provides symbols defined in the tool document."""

    def get_document_symbols(self, xml_document: XmlDocument) -> List[DocumentSymbol]:
        """Gets all symbols defined in the tool document in a hierarchical structure."""
        if xml_document.root is None:
            return []
        return [self._get_element_symbol_definition(xml_document, xml_document.root)]

    def _get_element_children_symbols(self, element: XmlElement, xml_document: XmlDocument) -> List[DocumentSymbol]:
        result: List[DocumentSymbol] = []
        for child in element.children:
            if isinstance(child, XmlAttribute):
                result.append(self._get_attribute_symbol_definition(xml_document, child))
            if isinstance(child, XmlElement):
                result.append(self._get_element_symbol_definition(xml_document, child))
        return result

    def _get_element_symbol_definition(self, xml_document: XmlDocument, element: XmlElement) -> DocumentSymbol:
        element_range = convert_document_offsets_to_range(xml_document.document, element.start, element.end)
        return DocumentSymbol(
            name=self._get_node_name(element),
            kind=SymbolKind.Field,
            detail=self._get_element_symbol_detail(element, xml_document),
            range=element_range,
            selection_range=element_range,
            children=self._get_element_children_symbols(element, xml_document),
        )

    def _get_attribute_symbol_definition(self, xml_document: XmlDocument, attribute: XmlAttribute) -> DocumentSymbol:
        attribute_range = convert_document_offsets_to_range(xml_document.document, attribute.start, attribute.end)
        return DocumentSymbol(
            name=self._get_node_name(attribute),
            kind=SymbolKind.Property,
            detail=attribute.value.unquoted if attribute.value else None,
            range=attribute_range,
            selection_range=attribute_range,
        )

    def _get_node_name(self, node: XmlSyntaxNode) -> str:
        return node.name or ""

    def _get_element_symbol_detail(self, element: XmlElement, xml_document: XmlDocument) -> Optional[str]:
        if element.name in ["option", "when", "add", "remove"]:
            detail = element.get_attribute_value("value")
        elif element.name in ["citation", "validator"]:
            detail = element.get_attribute_value("type")
        elif element.name in ["requirement", "import"]:
            detail = element.get_content(xml_document.document.source)
        elif element.name in ["expand"]:
            detail = element.get_attribute_value("macro")
        else:
            detail = element.get_attribute_value("id") or element.get_attribute_value("name")
        return detail



"""This module contains definitions for constants used by the Galaxy Tools Language Server."""
from lsprotocol.types import (
    Position,
    Range,
)


class Commands:
    AUTO_CLOSE_TAGS = "gls.completion.autoCloseTags"
    GENERATE_TESTS = "gls.generate.tests"
    GENERATE_COMMAND = "gls.generate.command"
    SORT_SINGLE_PARAM_ATTRS = "gls.sort.singleParamAttributes"
    SORT_DOCUMENT_PARAMS_ATTRS = "gls.sort.documentParamsAttributes"
    DISCOVER_TESTS_IN_WORKSPACE = "gls.tests.discoverInWorkspace"
    DISCOVER_TESTS_IN_DOCUMENT = "gls.tests.discoverInDocument"
    GENERATE_EXPANDED_DOCUMENT = "gls.generate.expandedDocument"


class DiagnosticCodes:
    INVALID_EXPANDED_TOOL = 101


# Default Document Range and the start of the document
DEFAULT_DOCUMENT_RANGE = Range(
    start=Position(line=0, character=0),
    end=Position(line=0, character=0),
)



NEW_LINE = "\n"

_EXL = ord("!")
_MIN = ord("-")
_UDS = ord("_")
_LAN = ord("<")
_RAN = ord(">")
_FSL = ord("/")
_EQS = ord("=")
_DQO = ord('"')
_SQO = ord("'")
_NWL = ord(NEW_LINE)
_CAR = ord("\r")
_LFD = ord("\f")
_WSP = ord(" ")
_TAB = ord("\t")
_OSB = ord("[")
_CSB = ord("]")
_CVL = ord("C")
_DVL = ord("D")
_AVL = ord("A")
_TVL = ord("T")
_QMA = ord("?")

WHITESPACE_CHARS = [_WSP, _TAB, _NWL, _LFD, _CAR]
QUOTE_CHARS = [_DQO, _SQO]
COMMENT_END_CHAR_SEQ = [_MIN, _MIN, _RAN]
COMMENT_START_CHAR_SEQ = [_EXL, _MIN, _MIN]
CDATA_START_CHAR_SEQ = [_EXL, _OSB, _CVL, _DVL, _AVL, _TVL, _AVL, _OSB]
CDATA_END_CHAR_SEQ = [_CSB, _CSB, _RAN]
PI_END_CHAR_SEQ = [_QMA, _RAN]

UNDEFINED_OFFSET = -1



""" This code is based on the Eclipse/Lemminx XML language server implementation:
https://github.com/eclipse/lemminx/tree/master/org.eclipse.lemminx/src/main/java/org/eclipse/lemminx/dom

Only the minimum subset of the XML dialect used by Galaxy tool wrappers is supported.
"""

from typing import Optional

from .constants import (
    _EQS,
    _EXL,
    _FSL,
    _LAN,
    _QMA,
    _RAN,
    _UDS,
    CDATA_END_CHAR_SEQ,
    CDATA_START_CHAR_SEQ,
    COMMENT_END_CHAR_SEQ,
    COMMENT_START_CHAR_SEQ,
    PI_END_CHAR_SEQ,
    QUOTE_CHARS,
)
from .types import (
    ScannerState,
    TokenType,
)
from .utils import MultiLineStream

ERROR_UNEXPECTED_WHITESPACE = "Unexpected whitespace. Tag name must directly follow the open angle bracket."


class XmlScanner:
    """This class allows to sequentially scan a XML document to find and extract the exact positions of every
    token inside the document."""

    def __init__(self, source: str, initial_offset: int = 0, initial_state: ScannerState = ScannerState.WithinContent) -> None:
        self.stream = MultiLineStream(source, initial_offset)
        self.state = initial_state
        self.token_offset = 0
        self.token_type = TokenType.Unknown
        self.token_error: Optional[str] = None

    def scan(self) -> TokenType:
        """Scans the document to sequentially find the next token."""
        ofsset = self.stream.pos()
        token = self._internal_scan()
        if token != TokenType.EOS and ofsset == self.stream.pos():
            self.stream.advance(1)
            return self._finish_token(ofsset, TokenType.Unknown)
        return token

    def get_token_offset(self) -> int:
        """Gets the current token offset inside the document."""
        return self.token_offset

    def get_token_end(self) -> int:
        """Gets the last position/offset of the token."""
        return self.stream.pos()

    def get_token_text(self) -> str:
        """Gets the text of this token from its offset to the current position."""
        return self.stream.get_source()[self.token_offset : self.stream.pos()]

    def get_token_text_from_offset(self, offset: int) -> str:
        """Gets the text of this token from the given offset to the current position."""
        return self.stream.get_source()[offset : self.stream.pos()]

    def _finish_token(self, offset: int, type: TokenType, error_message: Optional[str] = None) -> TokenType:
        self.token_type = type
        self.token_offset = offset
        self.token_error = error_message
        return type

    # flake8: noqa: C901
    def _internal_scan(self) -> TokenType:
        """Scans the document for the next token.

        This method is a bit too complex, but, since it is a Python translation
        from the Java Eclipse/Lemminx parser, it could be easier to maintain it this way.

        Returns:
            TokenType: The token found.
        """
        offset = self.stream.pos()
        if self.stream.eos():
            return self._finish_token(offset, TokenType.EOS)

        if self.state == ScannerState.WithinComment:
            if self.stream.advance_if_chars(COMMENT_END_CHAR_SEQ):
                self.state = ScannerState.WithinContent
                return self._finish_token(offset, TokenType.EndCommentTag)
            self.stream.advance_until_chars(COMMENT_END_CHAR_SEQ)
            return self._finish_token(offset, TokenType.Comment)

        elif self.state == ScannerState.PrologOrPI:
            if self.stream.advance_if_chars(PI_END_CHAR_SEQ):
                self.state = ScannerState.WithinContent
                return self._finish_token(offset, TokenType.PIEnd)
            self.stream.advance_until_chars_or_new_tag(PI_END_CHAR_SEQ)
            if self.stream.peek_char() == _LAN:
                self.state = ScannerState.WithinContent
            return self._internal_scan()

        elif self.state == ScannerState.WithinPI:
            if self.stream.skip_whitespace():
                return self._finish_token(offset, TokenType.Whitespace)
            if self.stream.advance_if_chars(PI_END_CHAR_SEQ):
                self.state = ScannerState.WithinContent
                return self._finish_token(offset, TokenType.PIEnd)
            if self.stream.advance_until_chars_or_new_tag(PI_END_CHAR_SEQ):
                if self.stream.peek_char() == _LAN:
                    self.state = ScannerState.WithinContent
                if len(self.get_token_text_from_offset(offset)) == 0:
                    return self._finish_token(offset, TokenType.PIEnd)
            return self._finish_token(offset, TokenType.PIContent)

        elif self.state == ScannerState.WithinContent:
            if self.stream.advance_if_char(_LAN):
                if not self.stream.eos() and self.stream.peek_char() == _EXL:
                    if self.stream.advance_if_chars(CDATA_START_CHAR_SEQ):
                        self.state = ScannerState.WithinCDATA
                        return self._finish_token(offset, TokenType.CDATATagOpen)
                    if self.stream.advance_if_chars(COMMENT_START_CHAR_SEQ):
                        self.state = ScannerState.WithinComment
                        return self._finish_token(offset, TokenType.StartCommentTag)
                elif not self.stream.eos() and self.stream.peek_char() == _QMA:
                    self.state = ScannerState.PrologOrPI
                    return self._finish_token(offset, TokenType.StartPrologOrPI)
                if self.stream.advance_if_char(_FSL):
                    self.state = ScannerState.AfterOpeningEndTag
                    return self._finish_token(offset, TokenType.EndTagOpen)
                self.state = ScannerState.AfterOpeningStartTag
                return self._finish_token(offset, TokenType.StartTagOpen)
            self.stream.advance_until_char(_LAN)
            return self._finish_token(offset, TokenType.Content)

        elif self.state == ScannerState.WithinCDATA:
            if self.stream.advance_if_chars(CDATA_END_CHAR_SEQ):
                self.state = ScannerState.WithinContent
                return self._finish_token(offset, TokenType.CDATATagClose)
            self.stream.advance_until_chars(CDATA_END_CHAR_SEQ)
            return self._finish_token(offset, TokenType.CDATAContent)

        elif self.state == ScannerState.AfterOpeningEndTag:
            if self._has_next_element_name():
                self.state = ScannerState.WithinEndTag
                return self._finish_token(offset, TokenType.EndTag)
            if self.stream.skip_whitespace():
                return self._finish_token(
                    offset,
                    TokenType.Whitespace,
                    ERROR_UNEXPECTED_WHITESPACE,
                )
            self.state = ScannerState.WithinEndTag
            if self.stream.advance_until_char_or_new_tag(_RAN):
                if self.stream.peek_char() == _RAN:
                    self.state = ScannerState.WithinContent
                return self._internal_scan()
            return self._finish_token(offset, TokenType.Unknown)

        elif self.state == ScannerState.WithinEndTag:
            if self.stream.skip_whitespace():
                return self._finish_token(offset, TokenType.Whitespace)
            if self.stream.advance_if_char(_RAN):
                self.state = ScannerState.WithinContent
                return self._finish_token(offset, TokenType.EndTagClose)
            if self.stream.advance_until_char(_LAN):
                self.state = ScannerState.WithinContent
                return self._internal_scan()
            return self._finish_token(offset, TokenType.Whitespace)

        elif self.state == ScannerState.AfterOpeningStartTag:
            if self._has_next_element_name():
                self.state = ScannerState.WithinTag
                return self._finish_token(offset, TokenType.StartTag)
            if self.stream.skip_whitespace():
                return self._finish_token(
                    offset,
                    TokenType.Whitespace,
                    ERROR_UNEXPECTED_WHITESPACE,
                )
            self.state = ScannerState.WithinTag
            if self.stream.advance_until_char_or_new_tag(_RAN):
                if self.stream.peek_char() == _LAN:
                    self.state = ScannerState.WithinContent
                return self._internal_scan()
            return self._finish_token(offset, TokenType.Unknown)

        elif self.state == ScannerState.WithinTag:
            if self.stream.skip_whitespace():
                return self._finish_token(offset, TokenType.Whitespace)
            if self.stream.advance_if_chars(PI_END_CHAR_SEQ):
                self.state = ScannerState.WithinContent
                return self._finish_token(offset, TokenType.PIEnd)
            if self._has_next_attribute_name():
                self.state = ScannerState.AfterAttributeName
                return self._finish_token(offset, TokenType.AttributeName)
            if self.stream.advance_if_char(_FSL):
                self.state = ScannerState.WithinTag
                if self.stream.advance_if_char(_RAN):
                    self.state = ScannerState.WithinContent
                    return self._finish_token(offset, TokenType.StartTagSelfClose)
                return self._finish_token(offset, TokenType.Unknown)
            ch = self.stream.peek_char()
            if ch in QUOTE_CHARS:
                self.state = ScannerState.BeforeAttributeValue
                return self._internal_scan()
            if self.stream.advance_if_char(_RAN):
                self.state = ScannerState.WithinContent
                return self._finish_token(offset, TokenType.StartTagClose)
            if self.stream.advance_until_char(_LAN):
                self.state = ScannerState.WithinContent
                return self._internal_scan()
            return self._finish_token(offset, TokenType.Unknown)

        elif self.state == ScannerState.AfterAttributeName:
            if self.stream.skip_whitespace():
                return self._finish_token(offset, TokenType.Whitespace)
            if self.stream.advance_if_char(_EQS):
                self.state = ScannerState.BeforeAttributeValue
                return self._finish_token(offset, TokenType.DelimiterAssign)
            self.state = ScannerState.WithinTag
            return self._internal_scan()

        elif self.state == ScannerState.BeforeAttributeValue:
            if self.stream.skip_whitespace():
                return self._finish_token(offset, TokenType.Whitespace)
            if self._has_next_attribute_value():
                self.state = ScannerState.WithinTag
                return self._finish_token(offset, TokenType.AttributeValue)
            self.state = ScannerState.WithinTag
            return self._internal_scan()

        self.stream.advance(1)
        self.state = ScannerState.WithinContent
        return self._finish_token(offset, TokenType.Unknown)

    def _has_next_element_name(self) -> bool:
        first = self.stream.peek_char()
        if not self._is_valid_start_name_character(chr(first)):
            return False
        self.stream.advance(1)
        self.stream.advance_while_char(self._is_valid_name_character)
        return True

    def _has_next_attribute_name(self) -> bool:
        return self.stream.advance_while_char(self._is_valid_name_character) > 0

    def _has_next_attribute_value(self) -> bool:
        first = self.stream.peek_char()
        if first in QUOTE_CHARS:
            self.stream.advance(1)
            if self.stream.advance_until_char(first):
                self.stream.advance(1)
            return True
        return False

    def _is_valid_name_character(self, ch: str) -> bool:
        return ord(ch) == _UDS or ch.isalnum()

    def _is_valid_start_name_character(self, ch: str) -> bool:
        return ord(ch) == _UDS or ch.isalpha()  # No numbers allowed as first character



"""Utilities to validate Galaxy xml tool wrappers and extract
information from the XSD schema.
"""

from typing import (
    List,
    Optional,
)

from lsprotocol.types import (
    Diagnostic,
    MarkupContent,
    MarkupKind,
)
from lxml import etree

from galaxyls.services.context import XmlContext
from galaxyls.services.xml.document import XmlDocument
from galaxyls.services.xsd.constants import (
    MSG_NO_DOCUMENTATION_AVAILABLE,
    TOOL_XSD_FILE,
)
from galaxyls.services.xsd.parser import GalaxyToolXsdParser
from galaxyls.services.xsd.types import XsdBase
from galaxyls.services.xsd.validation import GalaxyToolSchemaValidationService

NO_DOC_MARKUP = MarkupContent(kind=MarkupKind.Markdown, value=MSG_NO_DOCUMENTATION_AVAILABLE)


class GalaxyToolXsdService:
    """Galaxy tool Xml Schema Definition service.

    This service provides functionality to extract information from
    the XSD schema and validate XML files against it.
    """

    def __init__(self) -> None:
        """Initializes the validator by loading the XSD."""
        self.xsd_doc: etree._ElementTree = etree.parse(str(TOOL_XSD_FILE))
        self.xsd_schema = etree.XMLSchema(self.xsd_doc)
        self.xsd_parser = GalaxyToolXsdParser(self.xsd_doc.getroot())
        self.validator = GalaxyToolSchemaValidationService(self.xsd_schema)

    def validate_document(self, xml_document: XmlDocument) -> List[Diagnostic]:
        """Validates the Galaxy tool xml using the XSD schema and returns a list
        of diagnostics if there are any problems.
        """
        return self.validator.validate_document(xml_document)

    def get_documentation_for(self, context: XmlContext) -> MarkupContent:
        """Gets the documentation annotated in the XSD about the
        given element name (node or attribute).
        """
        if context.xsd_element:
            element: Optional[XsdBase]
            if context.is_tag:
                element = context.xsd_element
            elif context.is_attribute_key and context.node and context.node.name:
                element = context.xsd_element.attributes.get(context.node.name)
            if element:
                return element.get_doc()
        return NO_DOC_MARKUP



from typing import (
    List,
    Optional,
)

from lsprotocol.types import (
    CodeAction,
    CodeActionParams,
    CompletionList,
    CompletionParams,
    Diagnostic,
    DocumentFormattingParams,
    Hover,
    Location,
    MarkupContent,
    MarkupKind,
    Position,
    Range,
    TextEdit,
)
from pygls.workspace import (
    Document,
    Workspace,
)

from galaxyls.services.definitions import DocumentDefinitionsProvider
from galaxyls.services.links import DocumentLinksProvider
from galaxyls.services.macros import MacroExpanderService
from galaxyls.services.symbols import DocumentSymbolsProvider
from galaxyls.services.tools.common import (
    TestsDiscoveryService,
    ToolParamAttributeSorter,
)
from galaxyls.services.tools.document import GalaxyToolXmlDocument
from galaxyls.services.tools.generators.command import GalaxyToolCommandSnippetGenerator
from galaxyls.services.tools.generators.tests import GalaxyToolTestSnippetGenerator
from galaxyls.services.tools.iuc import IUCToolParamAttributeSorter
from galaxyls.services.tools.linting import GalaxyToolLinter
from galaxyls.services.tools.macros import MacroDefinitionsProvider
from galaxyls.services.tools.refactor import (
    RefactoringService,
    RefactorMacrosService,
)
from galaxyls.services.tools.testing import ToolTestsDiscoveryService
from ..config import CompletionMode
from ..types import (
    GeneratedSnippetResult,
    ReplaceTextRangeResult,
)
from .completion import (
    AutoCloseTagResult,
    XmlCompletionService,
)
from .context import XmlContextService
from .format import GalaxyToolFormatService
from .xml.document import XmlDocument
from .xsd.service import GalaxyToolXsdService


class GalaxyToolLanguageService:
    """Galaxy tool Language service.

    This service manages all the operations supported
    by the LSP.
    """

    def __init__(self) -> None:
        self.xsd_service = GalaxyToolXsdService()
        self.format_service = GalaxyToolFormatService()
        self.xsd_tree = self.xsd_service.xsd_parser.get_tree()
        self.xml_context_service = XmlContextService(self.xsd_tree)
        self.sort_service: ToolParamAttributeSorter = IUCToolParamAttributeSorter()
        self.test_discovery_service: TestsDiscoveryService = ToolTestsDiscoveryService()
        self.macro_expander = MacroExpanderService()
        self.refactoring_service: Optional[RefactoringService] = None
        self.linter = GalaxyToolLinter()
        self.definitions_provider: Optional[DocumentDefinitionsProvider] = None
        self.link_provider = DocumentLinksProvider()
        self.symbols_provider = DocumentSymbolsProvider()

    def set_workspace(self, workspace: Workspace) -> None:
        macro_definitions_provider = MacroDefinitionsProvider(workspace)
        self.definitions_provider = DocumentDefinitionsProvider(macro_definitions_provider)
        self.completion_service = XmlCompletionService(self.xsd_tree, self.definitions_provider)
        self.refactoring_service = RefactoringService(
            RefactorMacrosService(workspace, macro_definitions_provider, self.format_service)
        )

    def get_diagnostics(self, xml_document: XmlDocument) -> List[Diagnostic]:
        """Validates the Galaxy tool XML document and returns a list of diagnostics if there are any problems."""
        return self.xsd_service.validate_document(xml_document) + self.linter.lint_document(xml_document)

    def get_documentation(self, xml_document: XmlDocument, position: Position) -> Optional[Hover]:
        """Gets the documentation about the element at the given position."""
        context = self.xml_context_service.get_xml_context(xml_document, position)
        if context.node:
            if context.is_tag or context.is_attribute_key:
                documentation = self.xsd_service.get_documentation_for(context)
                context_range = self.xml_context_service.get_range_for_context(xml_document, context)
                return Hover(contents=documentation, range=context_range)
            # Try to get token
            word = xml_document.document.word_at_position(position)
            if self.definitions_provider:
                token = self.definitions_provider.get_token_definition(xml_document, word)
                if token:
                    return Hover(
                        contents=MarkupContent(kind=MarkupKind.Markdown, value=token.value),
                        range=Range(start=position, end=position),
                    )
        return None

    def format_document(self, content: str, params: DocumentFormattingParams) -> List[TextEdit]:
        """Given the document contents returns the list of TextEdits needed to properly format and layout the document."""
        return self.format_service.format(content, params)

    def get_completion(
        self, xml_document: XmlDocument, params: CompletionParams, mode: CompletionMode
    ) -> Optional[CompletionList]:
        """Gets completion items depending on the current document context."""
        if params.context is None:
            return None
        context = self.xml_context_service.get_xml_context(xml_document, params.position)
        return self.completion_service.get_completion_at_context(context, params.context, mode)

    def get_auto_close_tag(self, xml_document: XmlDocument, position: Position) -> Optional[AutoCloseTagResult]:
        """Gets the closing result for the currently opened tag in context."""
        # The trigger character `/` or `>` is placed right before the actual position, so we get the position.character - 1
        trigger_character = xml_document.document.lines[position.line][position.character - 1]
        # We want to get the context information right before the trigger character so we get position.character - 2
        position_before_trigger = Position(line=position.line, character=position.character - 2)
        context = self.xml_context_service.get_xml_context(xml_document, position_before_trigger)
        return self.completion_service.get_auto_close_tag(context, trigger_character)

    def generate_tests(self, document: Document) -> Optional[GeneratedSnippetResult]:
        """Generates a code snippet with some tests for the current inputs and outputs
        of this tool wrapper."""
        tool = GalaxyToolXmlDocument(document)
        generator = GalaxyToolTestSnippetGenerator(tool)
        return generator.generate_snippet()

    def generate_command(self, document: Document) -> Optional[GeneratedSnippetResult]:
        """Generates a boilerplate Cheetah code snippet based on the current inputs and outputs
        of this tool wrapper."""
        tool = GalaxyToolXmlDocument(document)
        generator = GalaxyToolCommandSnippetGenerator(tool)
        return generator.generate_snippet()

    def sort_single_param_attrs(self, xml_document: XmlDocument, position: Position) -> Optional[ReplaceTextRangeResult]:
        """Sorts the attributes of the param element under the cursor."""
        offset = xml_document.document.offset_at_position(position)
        param_element = xml_document.find_element_at(offset)
        if param_element:
            return self.sort_service.sort_param_attributes(param_element, xml_document)
        return None

    def sort_document_param_attributes(self, xml_document: XmlDocument) -> List[ReplaceTextRangeResult]:
        """Sorts the attributes of all the param elements contained in the document."""
        return self.sort_service.sort_document_param_attributes(xml_document)

    def get_available_refactoring_actions(
        self, xml_document: XmlDocument, params: CodeActionParams
    ) -> Optional[List[CodeAction]]:
        if self.refactoring_service is None:
            return None
        return self.refactoring_service.get_available_refactoring_actions(xml_document, params)

    def go_to_definition(self, xml_document: XmlDocument, position: Position) -> Optional[List[Location]]:
        if self.definitions_provider:
            return self.definitions_provider.go_to_definition(xml_document, position)
        return None



import abc
from typing import (
    List,
    Optional,
)

from lsprotocol.types import Diagnostic
from pygls.workspace import Workspace

from galaxyls.services.xml.document import XmlDocument
from galaxyls.services.xml.nodes import XmlElement
from galaxyls.types import (
    ReplaceTextRangeResult,
    TestSuiteInfoResult,
)


class ToolParamAttributeSorter(metaclass=abc.ABCMeta):
    """Interface to sort attributes inside an element and return them as a document replace range edit.

    The sort criteria is defined by the implementing class.
    """

    @abc.abstractmethod
    def sort_param_attributes(self, param: XmlElement, xml_document: XmlDocument) -> Optional[ReplaceTextRangeResult]:
        """Returns a document replace edit with the sorted attributes of the given param contained in the xml_document."""
        raise NotImplementedError

    @abc.abstractmethod
    def sort_document_param_attributes(self, xml_document: XmlDocument) -> List[ReplaceTextRangeResult]:
        """Returns a collection of edits with all the attributes of the param elements in the document sorted."""
        raise NotImplementedError


class TestsDiscoveryService(metaclass=abc.ABCMeta):
    """Interface class for test discovering."""

    @abc.abstractmethod
    def discover_tests_in_workspace(self, workspace: Workspace) -> List[TestSuiteInfoResult]:
        raise NotImplementedError

    @abc.abstractmethod
    def discover_tests_in_document(self, xml_document: XmlDocument) -> Optional[TestSuiteInfoResult]:
        raise NotImplementedError


class ToolLinter(metaclass=abc.ABCMeta):
    """Interface class for linting tool documents."""

    @abc.abstractmethod
    def lint_document(self, xml_document: XmlDocument) -> List[Diagnostic]:
        """Returns a list of diagnostics with all the issues found in the given document."""
        raise NotImplementedError



from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    cast,
    List,
    Optional,
    Tuple,
    Union,
)

from galaxy.util import xml_macros
from lsprotocol.types import (
    Position,
    Range,
)
from lxml import etree
from pygls.workspace import Document

from galaxyls.services.tools.constants import (
    DASH,
    UNDERSCORE,
)
from galaxyls.services.tools.document import GalaxyToolXmlDocument
from galaxyls.types import GeneratedSnippetResult


class SnippetGenerator(ABC):
    """This abstract class defines an XML code snippet generator that can create TextMate
    snippets using the information of the tool document."""

    def __init__(self, tool_document: GalaxyToolXmlDocument, tabSize: int = 4) -> None:
        self.tool_document = tool_document
        self.expanded_document = self._get_expanded_tool_document(tool_document)
        self.tabstop_count: int = 0
        self.indent_spaces: str = " " * tabSize
        super().__init__()

    def generate_snippet(self) -> GeneratedSnippetResult:
        """Generates a code snippet using this generator."""
        result, is_error = self._build_snippet()
        if is_error:
            return GeneratedSnippetResult.as_error(result)
        insert_position = self._find_snippet_insert_position()
        if type(insert_position) == Range:
            insert_position = cast(Range, insert_position)
            return GeneratedSnippetResult(result, insert_position.start, insert_position)
        insert_position = cast(Position, insert_position)
        return GeneratedSnippetResult(result, insert_position)

    @abstractmethod
    def _build_snippet(self) -> Tuple[str, bool]:
        """This abstract function should return a tuple with the generated snippet text in TextMate format or
        an error message if the snippet can't be generated.

        The second value of the tuple is a bool indicating if there was an error."""
        pass

    @abstractmethod
    def _find_snippet_insert_position(self) -> Union[Position, Range]:
        """This abstract function should find the proper position inside the document where the
        snippet will be inserted."""
        pass

    def _get_expanded_tool_document(self, tool_document: GalaxyToolXmlDocument) -> GalaxyToolXmlDocument:
        """If the given tool document uses macros, a new tool document with the expanded macros is returned,
        otherwise, the same document is returned.
        """
        if tool_document.uses_macros:
            try:
                document = tool_document.document
                expanded_tool_tree, _ = xml_macros.load_with_references(document.path)
                expanded_tool_tree = cast(etree._ElementTree, expanded_tool_tree)
                expanded_source = etree.tostring(expanded_tool_tree, encoding=str)
                expanded_document = Document(uri=document.uri, source=expanded_source, version=document.version)
                return GalaxyToolXmlDocument(expanded_document)
            except BaseException:
                return tool_document
        return tool_document

    def _get_next_tabstop(self) -> str:
        """Increments the tabstop count and returns the current tabstop
        in TextMate format.

        Returns:
            str: The current tabstop for the code snippet.
        """
        self.tabstop_count += 1
        return f"${self.tabstop_count}"

    def _get_next_tabstop_with_placeholder(self, placeholder: str) -> str:
        """Returns the current tabstop with a placeholder text.

        Args:
            placeholder (str): The placeholder text that will appear on this tabstop.

        Returns:
            str: The current tabstop with the placeholder text.
        """
        self.tabstop_count += 1
        return f"${{{self.tabstop_count}:{placeholder}}}"

    def _get_next_tabstop_with_options(self, options: List[str], default_option: Optional[str] = None) -> str:
        """Gets the current tabstop with a list of possible options.

        If the list is empty, a normal tabstop is returned.

        Args:
            options (List[str]): The list of options that can be selected in this tabstop.

        Returns:
            str: The current tabstop with all the available options.
        """
        if options:
            if default_option:
                options.remove(default_option)
                options.insert(0, default_option)
            self.tabstop_count += 1
            return f"${{{self.tabstop_count}|{','.join(options)}|}}"
        return self._get_next_tabstop()

    def _extract_name_from_argument(self, argument: str) -> str:
        return argument.lstrip(DASH).replace(DASH, UNDERSCORE)



from typing import (
    NamedTuple,
    Type,
    TypeVar,
)

from lsprotocol import converters as cv

T = TypeVar("T")


def convert_to(params: NamedTuple, type: Type[T]) -> T:
    """Given a namedtuple, converts it into the given model type."""
    converter = cv.get_converter()
    obj = converter.structure(params, type)
    return obj



"""Module in charge of the auto-completion feature."""

from typing import (
    cast,
    List,
    Optional,
)

from lsprotocol.types import (
    CompletionContext,
    CompletionItem,
    CompletionItemKind,
    CompletionList,
    CompletionTriggerKind,
    InsertTextFormat,
    Position,
    Range,
)

from galaxyls.services.definitions import DocumentDefinitionsProvider
from galaxyls.services.xml.nodes import (
    XmlCDATASection,
    XmlElement,
)
from ..config import CompletionMode
from ..types import AutoCloseTagResult
from .context import XmlContext
from .xsd.parser import (
    XsdAttribute,
    XsdNode,
    XsdTree,
)


class XmlCompletionService:
    """Service in charge of generating completion lists based
    on the current XML context.
    """

    def __init__(self, xsd_tree: XsdTree, definitions_provider: DocumentDefinitionsProvider):
        self.xsd_tree: XsdTree = xsd_tree
        self.definitions_provider = definitions_provider

    def get_completion_at_context(
        self, context: XmlContext, completion_context: CompletionContext, mode: CompletionMode = CompletionMode.AUTO
    ) -> Optional[CompletionList]:
        if isinstance(context.node, XmlCDATASection):
            return None
        triggerKind = completion_context.trigger_kind
        if mode == CompletionMode.AUTO and triggerKind == CompletionTriggerKind.TriggerCharacter and not context.is_attribute:
            if completion_context.trigger_character == "<":
                return self.get_node_completion(context)
            if completion_context.trigger_character == " ":
                return self.get_attribute_completion(context)
        elif triggerKind == CompletionTriggerKind.Invoked:
            if context.is_inside_attribute_value:
                return self.get_attribute_value_completion(context)
            if context.is_attribute_key:
                return self.get_attribute_completion(context)
            if context.is_tag and not context.is_closing_tag and not context.is_at_end:
                if context.is_valid_tag() and not context.is_tag_name:
                    return self.get_attribute_completion(context)
                return self.get_node_completion(context)
        return None

    def get_node_completion(self, context: XmlContext) -> CompletionList:
        """Gets a list of completion items with all the available child tags
        that can be placed in the current context node.

        Args:
            context (XmlContext): The XML context information at a specific
            document position. It should contain, at least, the current node.

        Returns:
            CompletionList: A list of completion items with the child nodes
            that can be placed under the current node.
        """
        result = []
        if context.is_empty or context.is_root:
            result.append(self._build_node_completion_item(self.xsd_tree.root))
        elif context.xsd_element:
            for child in context.xsd_element.children:
                if not context.has_reached_max_occurs(child):
                    result.append(self._build_node_completion_item(child, len(result)))
            result.append(self._build_node_completion_item(self.xsd_tree.expand_element, len(result)))
        return CompletionList(items=result, is_incomplete=False)

    def get_attribute_completion(self, context: XmlContext) -> CompletionList:
        """Gets a list of completion items with all the attributes that can be
        used in the current context node.

        Args:
            context (XmlContext): The XML context information at a specific
            document position. It should contain, at least, the current node.

        Returns:
            CompletionList: The completion item with the basic information
            about the attributes.
        """
        result: List[CompletionItem] = []
        if (
            context.is_empty
            or context.is_content
            or context.is_attribute_value
            or context.is_closing_tag
            or not (context.node is not None and context.node.name)
        ):
            return CompletionList(items=result, is_incomplete=False)

        if context.xsd_element:
            existing_attr_names = context.node.get_attribute_names()
            for attr_name in context.xsd_element.attributes:
                if attr_name in existing_attr_names:
                    continue
                attr = context.xsd_element.attributes[attr_name]
                result.append(self._build_attribute_completion_item(attr, len(result)))
            if context.node.name == "expand":
                element = cast(XmlElement, context.node)
                macro_name = element.get_attribute_value("macro")
                if macro_name:
                    token_params = self.definitions_provider.macro_definitions_provider.get_macro_token_params(
                        context.xml_document, macro_name
                    )
                    for token in token_params:
                        if token.param_name in existing_attr_names:
                            continue
                        result.append(
                            CompletionItem(
                                label=token.param_name,
                                kind=CompletionItemKind.Variable,
                                insert_text=f'{token.param_name}="${{1:{token.default_value}}}"',
                                insert_text_format=InsertTextFormat.Snippet,
                                sort_text=str(len(result)).zfill(2),
                            )
                        )
        return CompletionList(items=result, is_incomplete=False)

    def get_attribute_value_completion(self, context: XmlContext) -> CompletionList:
        """Gets a list of possible values for an enumeration restricted attribute if exists.

        Args:
            context (XmlContext): The XML context at an attribute value position.

        Returns:
            CompletionList: The list of possible values of the attribute if it has an enumeration
            restriction.
        """
        if context.xsd_element and context.attribute_name:
            attribute = context.xsd_element.attributes.get(context.attribute_name)
            if attribute and attribute.enumeration:
                result = [CompletionItem(label=item, kind=CompletionItemKind.Value) for item in attribute.enumeration]
                return CompletionList(items=result, is_incomplete=False)
            if attribute and attribute.name == "macro":
                macro_names = self.definitions_provider.macro_definitions_provider.get_macro_names(context.xml_document)
                result = [CompletionItem(label=item, kind=CompletionItemKind.Value) for item in macro_names]
                return CompletionList(items=result, is_incomplete=False)
        return CompletionList(items=[], is_incomplete=False)

    def get_auto_close_tag(self, context: XmlContext, trigger_character: str) -> Optional[AutoCloseTagResult]:
        """Gets the closing result for the currently opened tag in context.

        The `context` parameter should be placed right before the trigger_character, otherwise the context
        information will be located at the trigger_character itself which doesn't provide the real context."""
        if (
            isinstance(context.node, XmlCDATASection)
            or context.is_closing_tag
            or (context.node is not None and context.node.is_closed)
            or (context.is_attribute and not context.is_attribute_end)
            or context.characted_at_position == ">"
            or context.xsd_element is None
        ):
            return None

        tag = context.xsd_element.name
        snippet = f"$0</{tag}>"
        replace_range = None
        is_self_closing = trigger_character == "/"
        if is_self_closing and context.position is not None:
            # Build the position Range to be replaced by the snippet
            # Get the document position of the trigger_character => +1 character from current context.position
            start = Position(line=context.position.line, character=context.position.character + 1)
            # Check if there is a `>` already after the `/` trigger and include it in the Range to avoid duplication
            end_character = context.position.character + 2
            if len(context.line_text) > end_character and context.line_text[end_character] == ">":
                end_character = end_character + 1
            end = Position(line=context.position.line, character=end_character)
            replace_range = Range(start=start, end=end)
            if not context.is_content:
                snippet = "/>$0"
        elif context.is_content:
            return None

        return AutoCloseTagResult(snippet, replace_range)

    def _build_node_completion_item(self, node: XsdNode, order: int = 0) -> CompletionItem:
        """Generates a completion item with the information about the
        given node definition.

        Args:
            node (XsdNode): The node definition used to build the
            completion item.
            order (int): The position for ordering this item.

        Returns:
            CompletionItem: The completion item with the basic information
            about the node.
        """
        return CompletionItem(
            label=node.name,
            kind=CompletionItemKind.Class,
            documentation=node.get_doc(),
            sort_text=str(order).zfill(2),
        )

    def _build_attribute_completion_item(self, attr: XsdAttribute, order: int = 0) -> CompletionItem:
        """Generates a completion item with the information about the
        given attribute definition.

        Args:
            attr (XsdAttribute): The attribute definition used to build the
            completion item.
            order (int): The position for ordering this item.

        Returns:
            CompletionItem: The completion item with the basic information
            about the attribute.
        """

        value_placeholder = "$1"
        if attr.enumeration:
            value_placeholder = f"${{1|{','.join(attr.enumeration)}|}}"
        return CompletionItem(
            label=attr.name,
            kind=CompletionItemKind.Variable,
            documentation=attr.get_doc(),
            insert_text=f'{attr.name}="{value_placeholder}"',
            insert_text_format=InsertTextFormat.Snippet,
            sort_text=str(order).zfill(2),
        )



import argparse
import logging

from .server import language_server

logging.basicConfig(filename="galaxy-language-server.log", level=logging.DEBUG, filemode="w")


def add_arguments(parser: argparse.ArgumentParser) -> None:
    parser.description = "Galaxy Language Server"

    parser.add_argument("--tcp", action="store_true", help="Use TCP server")
    parser.add_argument("--ws", action="store_true", help="Use WebSocket server")
    parser.add_argument("--host", default="127.0.0.1", help="Bind to this address")
    parser.add_argument("--port", type=int, default=2087, help="Bind to this port")


def main():
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    args = parser.parse_args()

    if args.tcp:
        language_server.start_tcp(args.host, args.port)
    elif args.ws:
        language_server.start_ws(args.host, args.port)
    else:
        language_server.start_io()


if __name__ == "__main__":
    main()



from typing import (
    cast,
    List,
)

from galaxy.tool_util.lint import (
    lint_xml_with,
    LintContext,
    LintLevel,
    LintMessage,
    XMLLintMessageXPath,
)
from lsprotocol.types import (
    Diagnostic,
    DiagnosticSeverity,
)

from galaxyls.services.tools.common import ToolLinter
from galaxyls.services.xml.document import XmlDocument


class GalaxyToolLinter(ToolLinter):
    diagnostics_source = "Galaxy Tool Linter"

    def lint_document(self, xml_document: XmlDocument) -> List[Diagnostic]:
        """Lint the given document using the Galaxy linter modules and return a list of Diagnostics."""
        result: List[Diagnostic] = []
        xml_tree = xml_document.xml_tree_expanded
        if not xml_document.is_tool_file or xml_tree is None:
            return result
        lint_context = LintContext(level=LintLevel.SILENT, lint_message_class=XMLLintMessageXPath)
        context = lint_xml_with(lint_context, xml_tree)
        result.extend(
            [
                self._to_diagnostic(lint_message, xml_document, DiagnosticSeverity.Error)
                for lint_message in context.error_messages
            ]
        )
        result.extend(
            [
                self._to_diagnostic(lint_message, xml_document, DiagnosticSeverity.Warning)
                for lint_message in context.warn_messages
            ]
        )
        return result

    def _to_diagnostic(self, lint_message: LintMessage, xml_document: XmlDocument, level: DiagnosticSeverity) -> Diagnostic:
        lint_message = cast(XMLLintMessageXPath, lint_message)
        range = xml_document.get_element_range_from_xpath_or_default(lint_message.xpath)
        result = Diagnostic(
            range=range,
            message=lint_message.message,
            source=self.diagnostics_source,
            severity=level,
        )
        return result



from enum import (
    Enum,
    unique,
)


@unique
class DocumentType(Enum):
    """Supported types of XML documents."""

    UNKNOWN = 1
    TOOL = 2
    MACROS = 3


@unique
class NodeType(Enum):
    """Supported types of document nodes."""

    UNKNOWN = 1
    ELEMENT = 2
    ATTRIBUTE = 3
    ATTRIBUTE_KEY = 4
    ATTRIBUTE_VALUE = 5
    CDATA_SECTION = 6
    COMMENT = 7
    CONTENT = 8
    PROCESSING_INSTRUCTION = 9
    DOCUMENT = 10


@unique
class TokenType(Enum):
    """Types of tokens in a XML document."""

    StartCommentTag = 1
    Comment = 2
    EndCommentTag = 3
    CDATATagOpen = 4
    CDATAContent = 5
    CDATATagClose = 6
    StartTagOpen = 7
    StartTagClose = 8
    StartTagSelfClose = 9
    StartTag = 10
    EndTagOpen = 11
    EndTagClose = 12
    EndTag = 13
    DelimiterAssign = 14
    AttributeName = 15
    AttributeValue = 16
    StartPrologOrPI = 17
    PrologName = 18
    PIName = 19
    PIContent = 20
    PIEnd = 21
    PrologEnd = 22
    Content = 23
    Whitespace = 24
    Unknown = 25
    EOS = 26


@unique
class ScannerState(Enum):
    """Possible states of the XML scanner."""

    WithinContent = 1
    AfterOpeningStartTag = 2
    AfterOpeningEndTag = 3
    WithinTag = 4
    WithinEndTag = 5
    WithinComment = 6
    AfterAttributeName = 7
    BeforeAttributeValue = 8
    WithinCDATA = 9
    AfterClosingCDATATag = 10
    StartCDATATag = 11
    AfterPrologOpen = 12
    PrologOrPI = 13
    WithinPI = 14



from pathlib import Path
from typing import (
    Dict,
    List,
    Optional,
    Set,
    Union,
)
from urllib.parse import urlparse

import attrs
from lsprotocol.types import (
    CodeAction,
    CodeActionKind,
    CodeActionParams,
    CreateFile,
    DeleteFile,
    OptionalVersionedTextDocumentIdentifier,
    Position,
    Range,
    RenameFile,
    ResourceOperationKind,
    TextDocumentEdit,
    TextEdit,
    WorkspaceEdit,
)
from lxml import etree
from pygls.workspace import Workspace

from galaxyls.services.format import GalaxyToolFormatService
from galaxyls.services.tools.constants import (
    DESCRIPTION,
    MACRO,
    MACROS,
    TOOL,
    XML,
    XREF,
)
from galaxyls.services.tools.document import GalaxyToolXmlDocument
from galaxyls.services.tools.macros import (
    ImportedMacrosFile,
    MacroDefinitionsProvider,
    ToolMacroDefinitions,
)
from galaxyls.services.xml.document import XmlDocument
from galaxyls.services.xml.nodes import XmlElement

DEFAULT_MACROS_FILENAME = "macros.xml"
EXCLUDED_TAGS = {TOOL, MACROS, MACRO, XML}

DocumentChanges = Optional[List[Union[TextDocumentEdit, CreateFile, RenameFile, DeleteFile]]]


@attrs.define
class MacroData:
    name: str
    content: str


class RefactorMacrosService:
    """Refactoring operations in the context of macros."""

    def __init__(
        self,
        workspace: Workspace,
        macro_definitions_provider: MacroDefinitionsProvider,
        format_service: GalaxyToolFormatService,
    ) -> None:
        self.workspace = workspace
        self.definitions_provider = macro_definitions_provider
        self.format_service = format_service

    def create_extract_to_local_macro_actions(
        self, tool: GalaxyToolXmlDocument, macro: MacroData, params: CodeActionParams
    ) -> List[CodeAction]:
        """Returns refactoring actions to extract a macro into the local <macros> section of a tool wrapper."""
        return [
            CodeAction(
                title="Extract to local macro",
                kind=CodeActionKind.RefactorExtract,
                edit=WorkspaceEdit(changes=self._calculate_local_changes_for_macro(tool, macro, params)),
            )
        ]

    def create_extract_to_macros_file_actions(
        self, tool: GalaxyToolXmlDocument, macro_definitions: ToolMacroDefinitions, macro: MacroData, params: CodeActionParams
    ) -> List[CodeAction]:
        """Returns refactoring actions for extracting a macro into an external macros definition file."""
        if not macro_definitions.imported_macros:
            return [
                CodeAction(
                    title=f"Extract to macro, create and import {DEFAULT_MACROS_FILENAME}",
                    kind=CodeActionKind.RefactorExtract,
                    edit=WorkspaceEdit(
                        document_changes=self._calculate_external_changes_for_macro_in_new_file(
                            tool, DEFAULT_MACROS_FILENAME, macro, params
                        )
                    ),
                )
            ]
        code_actions = []
        for file_name, macro_file_definition in macro_definitions.imported_macros.items():
            code_actions.append(
                CodeAction(
                    title=f"Extract to macro in {file_name}",
                    kind=CodeActionKind.RefactorExtract,
                    edit=WorkspaceEdit(
                        changes=self._calculate_external_changes_for_macro(tool, macro_file_definition, macro, params)
                    ),
                )
            )

        return code_actions

    def _calculate_local_changes_for_macro(
        self, tool: GalaxyToolXmlDocument, macro: MacroData, params: CodeActionParams
    ) -> Dict[str, List[TextEdit]]:
        """Returns a dictionary with the file uri and the TextEdit operations that will add a macro definition to
        the <macros> section of a tool wrapper. If the <macros> section doesn't exists it will also be created."""
        macros_element = tool.get_macros_element()
        edits: List[TextEdit] = []
        if macros_element is None:
            edits.append(self._edit_create_with_macros_section(tool, macro))
        else:
            edits.append(self._edit_add_macro_to_macros_section(tool, macros_element, macro))
        edits.append(self._edit_replace_range_with_macro_expand(tool, macro, params.range))
        changes = {params.text_document.uri: edits}
        return changes

    def _calculate_external_changes_for_macro(
        self,
        tool: GalaxyToolXmlDocument,
        macro_file_definition: ImportedMacrosFile,
        macro: MacroData,
        params: CodeActionParams,
    ) -> Dict[str, List[TextEdit]]:
        """Returns a dictionary with the document uri and the collection of TextEdit operations for that particular document.

        The edits will add the macro definition to the given imported macros file and replace the refactored macro with the
        corresponding <expand> element in the tool wrapper."""
        macros_xml_doc = macro_file_definition.document
        if macro_file_definition.file_uri is None or macros_xml_doc is None or macros_xml_doc.root is None:
            return {}
        macros_root = macros_xml_doc.root
        insert_position = macros_xml_doc.get_position_after_last_child(macros_root)
        insert_range = Range(start=insert_position, end=insert_position)
        macro_xml = f'<xml name="{macro.name}">\n{macro.content}\n</xml>'
        final_macro_xml = self._adapt_format(macros_xml_doc, insert_range, macro_xml)
        external_edit = TextEdit(
            range=insert_range,
            new_text=final_macro_xml,
        )
        changes = {
            params.text_document.uri: [self._edit_replace_range_with_macro_expand(tool, macro, params.range)],
            macro_file_definition.file_uri: [external_edit],
        }
        return changes

    def _calculate_external_changes_for_macro_in_new_file(
        self, tool: GalaxyToolXmlDocument, new_file_name: str, macro: MacroData, params: CodeActionParams
    ) -> DocumentChanges:
        """Returns a list of workspace document changes that will create a new macros.xml file with the given
        macro definition inside and also import the newly created file in the tool wrapper."""
        base_path = Path(urlparse(tool.xml_document.document.uri).path).parent
        new_file_uri = (base_path / new_file_name).as_uri()
        xml_content = f'<macros>\n<xml name="{macro.name}">\n{macro.content}\n</xml>\n</macros>'
        final_xml_content = self.format_service.format_content(xml_content)
        new_doc_insert_position = Position(line=0, character=0)
        tool_document = self.workspace.get_document(params.text_document.uri)
        changes: DocumentChanges = [
            CreateFile(uri=new_file_uri, kind=ResourceOperationKind.Create),
            TextDocumentEdit(
                text_document=OptionalVersionedTextDocumentIdentifier(
                    uri=new_file_uri,
                    version=0,
                ),
                edits=[
                    TextEdit(
                        range=Range(start=new_doc_insert_position, end=new_doc_insert_position),
                        new_text=final_xml_content,
                    ),
                ],
            ),
            TextDocumentEdit(
                text_document=OptionalVersionedTextDocumentIdentifier(
                    uri=tool_document.uri,
                    version=tool_document.version,
                ),
                edits=[
                    self._edit_create_import_macros_section(tool, DEFAULT_MACROS_FILENAME),
                    self._edit_replace_range_with_macro_expand(tool, macro, params.range),
                ],
            ),
        ]
        return changes

    def _edit_replace_range_with_macro_expand(self, tool: GalaxyToolXmlDocument, macro: MacroData, range: Range) -> TextEdit:
        """Returns the TextEdit operation that will replace the refactored macro with the corresponding <expand> element."""
        indentation = tool.xml_document.get_line_indentation(range.start.line)
        return TextEdit(
            range=self._get_range_from_line_start(range),
            new_text=f'{indentation}<expand macro="{macro.name}"/>',
        )

    def _edit_create_import_macros_section(self, tool: GalaxyToolXmlDocument, macros_file_name: str) -> TextEdit:
        """Returns the TextEdit operation that will add a macros file <import> definition to the existing
        <macros> section of a tool wrapper or also create the <macros> section if it doesn't exists."""
        macros_element = tool.find_element(MACROS)
        if macros_element:
            insert_position = tool.get_position_before_first_child(macros_element)
            macro_xml = f"<import>{macros_file_name}</import>"
        else:
            insert_position = self._find_macros_insert_position(tool)
            macro_xml = f"<macros>\n<import>{macros_file_name}</import>\n</macros>"
        insert_range = Range(start=insert_position, end=insert_position)
        final_macro_xml = self._adapt_format(tool.xml_document, insert_range, macro_xml)
        return TextEdit(
            range=insert_range,
            new_text=final_macro_xml,
        )

    def _edit_create_with_macros_section(self, tool: GalaxyToolXmlDocument, macro: MacroData) -> TextEdit:
        """Returns the TextEdit operation that will add a local <macros> section inside the tool wrapper along with
        the macro provided."""
        insert_position = self._find_macros_insert_position(tool)
        insert_range = Range(start=insert_position, end=insert_position)
        macro_xml = f'<macros>\n<xml name="{macro.name}">\n{macro.content}\n</xml>\n</macros>'
        final_macro_xml = self._adapt_format(tool.xml_document, insert_range, macro_xml)
        return TextEdit(
            range=insert_range,
            new_text=final_macro_xml,
        )

    def _edit_add_macro_to_macros_section(
        self, tool: GalaxyToolXmlDocument, macros_element: XmlElement, macro: MacroData
    ) -> TextEdit:
        """Returns the TextEdit operation that will add a macro definition to the <macros> section of a tool wrapper."""
        insert_position = tool.get_position_after_last_child(macros_element)
        insert_range = Range(start=insert_position, end=insert_position)
        macro_xml = f'<xml name="{macro.name}">\n{macro.content}\n</xml>'
        final_macro_xml = self._adapt_format(tool.xml_document, insert_range, macro_xml)
        return TextEdit(
            range=insert_range,
            new_text=final_macro_xml,
        )

    def _find_macros_insert_position(self, tool: GalaxyToolXmlDocument) -> Position:
        """Returns the position inside the document where the macros section
        can be inserted.

        Returns:
            Range: The position where the macros section can be inserted in the document.
        """
        section = tool.find_element(XREF)
        if section:
            return tool.get_position_after(section)
        section = tool.find_element(DESCRIPTION)
        if section:
            return tool.get_position_after(section)
        root = tool.find_element(TOOL)
        content_range = tool.get_content_range(root)
        if content_range:
            return content_range.start
        return Position(line=0, character=0)

    def _get_range_from_line_start(self, range: Range) -> Range:
        """Given an arbitrary document range, returns the same range but with the start offset always at the
        begining of the line."""
        return Range(start=Position(line=range.start.line, character=0), end=range.end)

    def _apply_indent(self, text: str, indent: str) -> str:
        """Applies the `indent` amount of spaces to all lines of the given text."""
        indented = indent + text.replace("\n", "\n" + indent)
        return indented

    def _adapt_format(
        self, xml_document: XmlDocument, insert_range: Range, xml_text: str, insert_in_new_line: bool = True
    ) -> str:
        """Adapts a chunk of XML text to the current formatting of the document given the insertion position."""
        formatted_macro = self.format_service.format_content(xml_text).rstrip()
        reference_line = insert_range.start.line
        if not insert_in_new_line:
            reference_line -= 1
        indent = xml_document.get_line_indentation(reference_line)
        final_macro_text = self._apply_indent(formatted_macro, indent)
        if insert_in_new_line:
            return f"\n{final_macro_text}"
        return final_macro_text


class RefactoringService:
    def __init__(self, macros_refactoring_service: RefactorMacrosService) -> None:
        self.macros = macros_refactoring_service

    def get_available_refactoring_actions(self, xml_document: XmlDocument, params: CodeActionParams) -> List[CodeAction]:
        """Gets a collection of possible refactoring code actions on a selected chunk of the document."""
        code_actions = []
        text_in_range = xml_document.get_text_in_range(params.range)
        target_element_tag = self._get_valid_full_element_tag(text_in_range)
        if target_element_tag is not None:
            macro = MacroData(name=target_element_tag, content=text_in_range.strip())
            macro_definitions = self.macros.definitions_provider.load_macro_definitions(xml_document)
            tool = GalaxyToolXmlDocument.from_xml_document(xml_document)
            code_actions.extend(self.macros.create_extract_to_macros_file_actions(tool, macro_definitions, macro, params))
            code_actions.extend(self.macros.create_extract_to_local_macro_actions(tool, macro, params))
        return code_actions

    def _get_valid_full_element_tag(self, xml_text: str) -> Optional[str]:
        """Given a chunk of XML text, returns the name of the tag inside it or None if the
        node is incomplete or sintactically wrong."""
        stripped_xml = xml_text.strip()
        if len(stripped_xml) < 5 or (stripped_xml[0] != "<" or stripped_xml[-1] != ">"):
            # Too short to be an element or doesn't look like an element
            return None
        return self._get_valid_node_tag(stripped_xml, EXCLUDED_TAGS)

    def _get_valid_node_tag(self, stripped_xml: str, exclude: Set[str]) -> Optional[str]:
        """Returns the name of the tag of a sintactically well formed xml text.

        The parameter `exclude` can define a set of tags that will be considered not valid."""
        try:
            xml_in_range = etree.fromstring(stripped_xml, etree.XMLParser(strip_cdata=False))
            if xml_in_range.tag not in exclude:
                return xml_in_range.tag
        except BaseException:
            pass  # Ignore, the xml chunk is not a valid xml node
        return None



import re
from typing import Optional

from pygls.workspace import Document

from galaxyls.services.xml.types import DocumentType

MAX_PEEK_CONTENT = 100
TAG_GROUP_NAME = "tag"
TAG_REGEX = rf"[\n\s]*?.*?[\n\s]*?<(?!\?)(?P<{TAG_GROUP_NAME}>[\w]*)"


class DocumentValidator:
    """Provides some utilities to quickly check documents without completely parse them beforehand."""

    @classmethod
    def has_valid_root(cls, document: Document) -> bool:
        """Checks if the document's root element matches one of the supported types
        or is an empty document."""
        if DocumentValidator.is_empty_document(document):
            return True
        root = DocumentValidator._get_document_root_tag(document)
        if root is not None:
            root_tag = root.upper()
            supported = [e.name for e in DocumentType if e != DocumentType.UNKNOWN]
            return root_tag == "" or root_tag in supported
        return False

    @classmethod
    def is_tool_document(cls, document: Document) -> bool:
        """Checks if the document's root element is <tool>."""
        root = DocumentValidator._get_document_root_tag(document)
        if root is not None:
            root_tag = root.upper()
            return root_tag == DocumentType.TOOL.name
        return False

    @classmethod
    def is_empty_document(cls, document: Document) -> bool:
        """Whether the document is empty or just contains spaces or empty lines."""
        return not document.source or document.source.isspace()

    @classmethod
    def _get_document_root_tag(cls, document: Document) -> Optional[str]:
        """Checks the first MAX_PEEK_CONTENT characters of the document for a root tag and
        returns the name of the tag if found."""
        content_peek = document.source[:MAX_PEEK_CONTENT]
        match = re.match(TAG_REGEX, content_peek)
        if match:
            group = match.group(TAG_GROUP_NAME)
            return group
        return None



from typing import (
    Dict,
    List,
    Optional,
    Set,
)

import attrs
from lsprotocol.types import Location
from pygls.workspace import Workspace

from galaxyls.services.tools.constants import (
    MACRO,
    NAME,
    TOKEN,
    XML,
)
from galaxyls.services.tools.document import GalaxyToolXmlDocument
from galaxyls.services.xml.document import XmlDocument
from galaxyls.services.xml.nodes import XmlElement
from galaxyls.services.xml.parser import XmlDocumentParser


@attrs.define
class TokenDefinition:
    name: str
    location: Location
    value: str


@attrs.define
class TokenParam(TokenDefinition):
    param_name: str
    default_value: str


@attrs.define
class MacroDefinition:
    name: str
    location: Location
    token_params: Dict[str, TokenParam]


@attrs.define
class ImportedMacrosFile:
    file_name: str
    file_uri: Optional[str]
    document: Optional[XmlDocument]
    tokens: Dict[str, TokenDefinition]
    macros: Dict[str, MacroDefinition]


@attrs.define
class ToolMacroDefinitions:
    tool_document: XmlDocument
    imported_macros: Dict[str, ImportedMacrosFile]
    tokens: Dict[str, TokenDefinition]
    macros: Dict[str, MacroDefinition]

    def go_to_import_definition(self, file_name: str) -> Optional[List[Location]]:
        imported_macros = self.imported_macros.get(file_name)
        if imported_macros and imported_macros.document and imported_macros.document.root and imported_macros.file_uri:
            macros_file_uri = imported_macros.file_uri
            content_range = imported_macros.document.get_full_range(imported_macros.document.root)
            if content_range:
                return [
                    Location(
                        uri=macros_file_uri,
                        range=content_range,
                    )
                ]
        return None

    def get_token_definition(self, token: str) -> Optional[TokenDefinition]:
        definition = self.tokens.get(token)
        if definition is None:
            return self.get_token_param_definition(token)
        return definition

    def get_token_param_definition(self, token: str) -> Optional[TokenParam]:
        for macro in self.macros.values():
            if token in macro.token_params:
                return macro.token_params.get(token)
        return None

    def get_macro_definition(self, macro_name: str) -> Optional[MacroDefinition]:
        return self.macros.get(macro_name)


class MacroDefinitionsProvider:
    """Provides location information about macros imported by a tool."""

    def __init__(self, workspace: Workspace) -> None:
        self.workspace = workspace

    def load_macro_definitions(self, tool_xml: XmlDocument) -> ToolMacroDefinitions:
        tool = GalaxyToolXmlDocument.from_xml_document(tool_xml)
        tokens = self._get_token_definitions(tool_xml)
        macros = self._get_macro_definitions(tool_xml)
        imported_macro_files = self._get_imported_macro_files_from_tool(tool)
        for file in imported_macro_files.values():
            tokens.update(file.tokens)
            macros.update(file.macros)
        return ToolMacroDefinitions(
            tool_document=tool_xml,
            imported_macros=imported_macro_files,
            tokens=tokens,
            macros=macros,
        )

    def get_macro_names(self, tool_xml: XmlDocument) -> Set[str]:
        tool = GalaxyToolXmlDocument.from_xml_document(tool_xml)
        macros = self._get_macro_definitions(tool_xml)
        imported_macro_files = self._get_imported_macro_files_from_tool(tool)
        for file in imported_macro_files.values():
            macros.update(file.macros)
        return set(macros.keys())

    def get_macro_token_params(self, tool_xml: XmlDocument, macro_name: str) -> List[TokenParam]:
        macro_definitions = self.load_macro_definitions(tool_xml)
        macro = macro_definitions.get_macro_definition(macro_name)
        if macro and macro.token_params:
            return list(macro.token_params.values())
        return []

    def _get_imported_macro_files_from_tool(self, tool: GalaxyToolXmlDocument) -> Dict[str, ImportedMacrosFile]:
        macro_files = {}
        uris_dict = tool.get_macro_import_uris()
        for file_name, file_uri in uris_dict.items():
            macros_document = self._load_macros_document(file_uri)
            tokens = self._get_token_definitions(macros_document)
            macros = self._get_macro_definitions(macros_document)
            macro_files[file_name] = ImportedMacrosFile(
                file_name=file_name,
                file_uri=file_uri,
                document=macros_document,
                tokens=tokens,
                macros=macros,
            )
        return macro_files

    def _load_macros_document(self, document_uri: str) -> XmlDocument:
        document = self.workspace.get_document(document_uri)
        xml_document = XmlDocumentParser().parse(document)
        return xml_document

    def _get_token_definitions(self, macros_xml: XmlDocument) -> Dict[str, TokenDefinition]:
        token_elements = macros_xml.find_all_elements_with_name(TOKEN)
        rval = {}
        for element in token_elements:
            token = element.get_attribute_value(NAME)
            if token is None:
                continue
            name = token.replace("@", "")
            value = element.get_content(macros_xml.document.source)
            token_def = TokenDefinition(
                name=name,
                location=Location(
                    uri=macros_xml.document.uri,
                    range=macros_xml.get_full_range(element) or XmlDocument.DEFAULT_RANGE,
                ),
                value=value,
            )
            rval[token_def.name] = token_def
        return rval

    def _get_macro_definitions(self, macros_xml: XmlDocument) -> Dict[str, MacroDefinition]:
        macro_elements = macros_xml.find_all_elements_with_name(MACRO)
        xml_elements = macros_xml.find_all_elements_with_name(XML)
        macro_elements += xml_elements
        rval = {}
        for element in macro_elements:
            name = element.get_attribute_value(NAME)
            macro_def = MacroDefinition(
                name=name or "Unknown",
                location=Location(
                    uri=macros_xml.document.uri,
                    range=macros_xml.get_full_range(element) or XmlDocument.DEFAULT_RANGE,
                ),
                token_params=self.get_token_params(macros_xml, element),
            )
            rval[macro_def.name] = macro_def
        return rval

    def get_token_params(self, macros_xml: XmlDocument, element: XmlElement) -> Dict[str, TokenParam]:
        token_params = {}
        for attr_name, attr in element.attributes.items():
            if attr_name.startswith("token_"):
                param_name = attr_name.replace("token_", "")
                token_name = param_name.upper()
                default_value = attr.get_value()
                token_param = TokenParam(
                    name=token_name,
                    param_name=param_name,
                    default_value=default_value,
                    value=f"**Token parameter**\n- Default value: `{default_value}`",
                    location=Location(
                        uri=macros_xml.document.uri,
                        range=macros_xml.get_full_range(attr) or XmlDocument.DEFAULT_RANGE,
                    ),
                )
                token_params[token_name] = token_param
        return token_params



from pathlib import Path
from typing import (
    cast,
    Dict,
    List,
    Optional,
)

from anytree import find
from lsprotocol.types import (
    Position,
    Range,
)
from pygls.workspace import Document

from galaxyls.services.tools.constants import (
    IMPORT,
    INPUTS,
    MACROS,
    OUTPUTS,
    TESTS,
    TOOL,
)
from galaxyls.services.tools.inputs import GalaxyToolInputTree
from galaxyls.services.xml.document import XmlDocument
from galaxyls.services.xml.nodes import (
    XmlContainerNode,
    XmlElement,
)
from galaxyls.services.xml.parser import XmlDocumentParser
from galaxyls.services.xml.types import DocumentType


class GalaxyToolXmlDocument:
    """Represents a Galaxy tool XML wrapper.

    This class provides access to the tool definitions and some utilities to extract
    information from the document.
    """

    def __init__(self, document: Document, xml_document: Optional[XmlDocument] = None) -> None:
        self.xml_document = xml_document or XmlDocumentParser().parse(document)
        self.document = document

    @property
    def is_valid(self) -> bool:
        """Indicates if this document is a valid Galaxy Tool Wrapper
        XML document."""
        return self.xml_document.document_type == DocumentType.TOOL

    @property
    def uses_macros(self) -> bool:
        """Indicates if this tool document *uses* macro definitions.

        Returns:
            bool: True if the tool contains <expand> elements.
        """
        return self.xml_document.uses_macros

    @property
    def source(self) -> str:
        """The contents of the tool document."""
        return self.xml_document.document.source

    @property
    def path(self) -> Optional[str]:
        """The file path of the tool."""
        return cast(Optional[str], self.xml_document.document.path)

    def find_element(self, name: str, maxlevel: int = 3) -> Optional[XmlElement]:
        """Finds the element with the given name in the document.

        Args:
            name (str): The name of the element to find.
            maxlevel (int, optional): The level at which the search will
            stop if not found. Defaults to 3.

        Returns:
            Optional[XmlElement]: The first element matching the name.
        """
        node = find(self.xml_document, filter_=lambda node: node.name == name, maxlevel=maxlevel)
        return cast(XmlElement, node)

    def get_content_range(self, element: Optional[XmlContainerNode]) -> Optional[Range]:
        """Returns the Range of the content block of the given element.

        Args:
            element (Optional[XmlElement]): The element.

        Returns:
            Optional[Range]: The Range of the content block.
        """
        if element:
            return self.xml_document.get_content_range(element)
        return None

    def get_position_before(self, element: XmlElement) -> Position:
        """Returns the document position right before the given element opening tag.

        Args:
            element (XmlElement): The element.

        Returns:
            Position: The position right before the element opening tag.
        """
        return self.xml_document.get_position_before(element)

    def get_position_after(self, element: XmlElement) -> Position:
        """Returns the document position right after the given element closing tag.

        Args:
            element (XmlElement): The element.

        Returns:
            Position: The position right after the element closing tag.
        """
        return self.xml_document.get_position_after(element)

    def get_position_after_last_child(self, element: XmlElement) -> Position:
        """Return the position in the document after the last child of the given element.

        Args:
            element (XmlElement): The element used to find the position.

        Returns:
            Position: The position just after the last child element declaration.
        """
        return self.xml_document.get_position_after_last_child(element)

    def get_position_before_first_child(self, element: XmlElement) -> Position:
        """Return the position in the document before the first child of the given element.

        Args:
            element (XmlElement): The element used to find the position.

        Returns:
            Position: The position just before the first child element declaration.
        """
        return self.xml_document.get_position_before_first_child(element)

    def analyze_inputs(self) -> GalaxyToolInputTree:
        """Gets the inputs in the document and builds the input tree.

        Returns:
            GalaxyToolInputTree: The resulting input tree for this document.
        """
        inputs = self.find_element(INPUTS)
        return GalaxyToolInputTree(inputs)

    def get_inputs(self) -> List[XmlElement]:
        """Gets the inputs of this document as a list of elements.

        Returns:
            List[XmlElement]: The outputs defined in the document.
        """
        outputs = self.find_element(INPUTS)
        if outputs:
            return outputs.elements
        return []

    def get_outputs(self) -> List[XmlElement]:
        """Gets the outputs of this document as a list of elements.

        Returns:
            List[XmlElement]: The outputs defined in the document.
        """
        outputs = self.find_element(OUTPUTS)
        if outputs:
            return outputs.elements
        return []

    def get_tool_element(self) -> Optional[XmlElement]:
        """Gets the root tool element"""
        return self.find_element(TOOL)

    def get_macros_element(self) -> Optional[XmlElement]:
        """Gets the macros element"""
        return self.find_element(MACROS)

    def get_macro_import_elements(self) -> Optional[List[XmlElement]]:
        """Gets the list of elements with macro file import declarations."""
        return self.xml_document.find_all_elements_with_name(IMPORT)

    def get_macro_import_uris(self) -> Dict[str, str]:
        """Returns a dictionary with the file name and corresponding file uri of each
        imported macro file."""
        result = {}
        tool_directory = self._get_tool_directory()
        import_elements = self.get_macro_import_elements()
        if import_elements:
            for imp in import_elements:
                filename = imp.get_content(self.xml_document.document.source)
                if filename:
                    path = tool_directory / filename
                    if path.exists():
                        file_uri = path.as_uri()
                        result[filename] = file_uri
        return result

    def get_macros_range(self) -> Optional[Range]:
        """Returns the Range position of the macros element name if it exists."""
        element = self.get_macros_element()
        if element:
            range = self.xml_document.get_element_name_range(element)
            return range
        return None

    def get_import_macro_file_range(self, file_path: Optional[str]) -> Optional[Range]:
        """Returns the Range position of the imported macro file element if it exists."""
        if file_path:
            filename = Path(file_path).name
            import_elements = self.get_macro_import_elements()
            if import_elements:
                for imp in import_elements:
                    imp_filename = imp.get_content(self.xml_document.document.source)
                    if imp_filename == filename:
                        return self.xml_document.get_full_range(imp)
        return None

    def get_tool_id(self) -> Optional[str]:
        """Gets the identifier of the tool"""
        tool_element = self.get_tool_element()
        return tool_element.get_attribute_value("id") if tool_element else None

    def get_tests(self) -> List[XmlElement]:
        """Gets the tests of this document as a list of elements.

        Returns:
            List[XmlElement]: The tests defined in the document.
        """
        tests = self.find_element(TESTS)
        if tests:
            return tests.elements
        return []

    def get_tests_range(self) -> Optional[Range]:
        """Returns the Range position of the tests element name if it exists."""
        element = self.find_element(TESTS)
        if element:
            range = self.xml_document.get_element_name_range(element)
            return range
        return None

    def get_test_data_path(self) -> Path:
        """Gets the test-data directory path of the tool"""
        tool_directory = self._get_tool_directory()
        return tool_directory / "test-data"

    @classmethod
    def from_xml_document(cls, xml_document: XmlDocument) -> "GalaxyToolXmlDocument":
        return GalaxyToolXmlDocument(xml_document.document, xml_document)

    def _get_tool_directory(self) -> Path:
        tool_directory = Path(self.xml_document.document.path).resolve().parent
        return tool_directory



"""Galaxy Tools Language Server implementation
"""
from typing import (
    List,
    Optional,
)

from lsprotocol.types import (
    CodeAction,
    CodeActionKind,
    CodeActionOptions,
    CodeActionParams,
    CompletionList,
    CompletionOptions,
    CompletionParams,
    ConfigurationItem,
    Diagnostic,
    DidChangeConfigurationParams,
    DidCloseTextDocumentParams,
    DidOpenTextDocumentParams,
    DidSaveTextDocumentParams,
    DocumentFormattingParams,
    DocumentLink,
    DocumentLinkParams,
    DocumentSymbol,
    DocumentSymbolParams,
    Hover,
    INITIALIZED,
    InitializeParams,
    Location,
    MessageType,
    TEXT_DOCUMENT_CODE_ACTION,
    TEXT_DOCUMENT_COMPLETION,
    TEXT_DOCUMENT_DEFINITION,
    TEXT_DOCUMENT_DID_CLOSE,
    TEXT_DOCUMENT_DID_OPEN,
    TEXT_DOCUMENT_DID_SAVE,
    TEXT_DOCUMENT_DOCUMENT_LINK,
    TEXT_DOCUMENT_DOCUMENT_SYMBOL,
    TEXT_DOCUMENT_FORMATTING,
    TEXT_DOCUMENT_HOVER,
    TextDocumentIdentifier,
    TextDocumentPositionParams,
    TextEdit,
    WORKSPACE_DID_CHANGE_CONFIGURATION,
    WorkspaceConfigurationParams,
)
from pygls.server import LanguageServer
from pygls.workspace import Document

from galaxyls.config import (
    CompletionMode,
    GalaxyToolsConfiguration,
)
from galaxyls.constants import Commands
from galaxyls.services.language import GalaxyToolLanguageService
from galaxyls.services.validation import DocumentValidator
from galaxyls.services.xml.document import XmlDocument
from galaxyls.services.xml.parser import XmlDocumentParser
from galaxyls.types import (
    AutoCloseTagResult,
    CommandParameters,
    GeneratedExpandedDocument,
    GeneratedSnippetResult,
    ReplaceTextRangeResult,
    TestSuiteInfoResult,
)
from galaxyls.utils import convert_to
from galaxyls.version import GLS_VERSION

GLS_NAME = "galaxy-tools-language-server"


class GalaxyToolsLanguageServer(LanguageServer):
    """Galaxy Tools Language Server."""

    def __init__(self) -> None:
        super().__init__(name=GLS_NAME, version=GLS_VERSION)
        self.service = GalaxyToolLanguageService()
        self.configuration: GalaxyToolsConfiguration = GalaxyToolsConfiguration()


language_server = GalaxyToolsLanguageServer()


async def _load_client_config_async(server: GalaxyToolsLanguageServer) -> None:
    """Loads the client configuration from user or workspace settings and updates
    the language server configuration.

    Args:
        server (GalaxyToolsLanguageServer): The language server instance.
    """
    try:
        config = await server.get_configuration_async(
            WorkspaceConfigurationParams(items=[ConfigurationItem(section="galaxyTools")])
        )
        server.configuration = convert_to(config[0], GalaxyToolsConfiguration)
    except BaseException as err:
        server.show_message_log(f"Error loading configuration: {err}")
        server.show_message("Error loading configuration. Using default settings.", MessageType.Error)


@language_server.feature(INITIALIZED)
async def initialized(server: GalaxyToolsLanguageServer, params: InitializeParams) -> None:
    """Loads the client configuration after initialization."""
    await _load_client_config_async(server)
    server.service.set_workspace(server.workspace)


@language_server.feature(WORKSPACE_DID_CHANGE_CONFIGURATION)
async def did_change_configuration(server: GalaxyToolsLanguageServer, params: DidChangeConfigurationParams) -> None:
    """Loads the client configuration after a change."""
    await _load_client_config_async(server)
    server.show_message("Settings updated")


@language_server.feature(TEXT_DOCUMENT_COMPLETION, CompletionOptions(trigger_characters=["<", " "]))
def completions(server: GalaxyToolsLanguageServer, params: CompletionParams) -> Optional[CompletionList]:
    """Returns completion items depending on the current document context."""
    if server.configuration.completion.mode == CompletionMode.DISABLED:
        return None
    document = _get_valid_document(server, params.text_document.uri)
    if document:
        xml_document = _get_xml_document(document)
        return server.service.get_completion(xml_document, params, server.configuration.completion.mode)
    return None


@language_server.feature(TEXT_DOCUMENT_HOVER)
def hover(server: GalaxyToolsLanguageServer, params: TextDocumentPositionParams) -> Optional[Hover]:
    """Displays Markdown documentation for the element under the cursor."""
    document = _get_valid_document(server, params.text_document.uri)
    if document:
        xml_document = _get_xml_document(document)
        return server.service.get_documentation(xml_document, params.position)
    return None


@language_server.feature(TEXT_DOCUMENT_FORMATTING)
def formatting(server: GalaxyToolsLanguageServer, params: DocumentFormattingParams) -> Optional[List[TextEdit]]:
    """Formats the whole document using the provided parameters"""
    document = _get_valid_document(server, params.text_document.uri)
    if document:
        content = document.source
        return server.service.format_document(content, params)
    return None


@language_server.feature(TEXT_DOCUMENT_DID_OPEN)
async def did_open(server: GalaxyToolsLanguageServer, params: DidOpenTextDocumentParams) -> None:
    """Occurs when a new xml document is open."""
    document = server.workspace.get_document(params.text_document.uri)
    if not DocumentValidator.is_empty_document(document):
        _validate(server, params)


@language_server.feature(TEXT_DOCUMENT_DID_SAVE)
def did_save(server: GalaxyToolsLanguageServer, params: DidSaveTextDocumentParams) -> None:
    """Occurs when the xml document is saved to disk."""
    _validate(server, params)


@language_server.feature(TEXT_DOCUMENT_DID_CLOSE)
def did_close(server: GalaxyToolsLanguageServer, params: DidCloseTextDocumentParams) -> None:
    """Occurs when the xml document is closed."""
    server.publish_diagnostics(params.text_document.uri, [])


@language_server.feature(TEXT_DOCUMENT_DEFINITION)
def definition(server: GalaxyToolsLanguageServer, params: TextDocumentPositionParams) -> Optional[List[Location]]:
    """Provides the location of a symbol definition."""
    document = _get_valid_document(server, params.text_document.uri)
    if document:
        xml_document = _get_xml_document(document)
        return server.service.go_to_definition(xml_document, params.position)
    return None


@language_server.feature(TEXT_DOCUMENT_DOCUMENT_LINK)
def document_link(server: GalaxyToolsLanguageServer, params: DocumentLinkParams) -> List[DocumentLink]:
    document = _get_valid_document(server, params.text_document.uri)
    if document:
        xml_document = _get_xml_document(document)
        return server.service.link_provider.get_document_links(xml_document)
    return []


@language_server.feature(
    TEXT_DOCUMENT_CODE_ACTION,
    CodeActionOptions(
        code_action_kinds=[
            CodeActionKind.RefactorExtract,
        ],
    ),
)
def process_code_actions(server: GalaxyToolsLanguageServer, params: CodeActionParams) -> Optional[List[CodeAction]]:
    document = _get_valid_document(server, params.text_document.uri)
    if document is None:
        return None
    xml_document = _get_xml_document(document)
    return server.service.get_available_refactoring_actions(xml_document, params)


@language_server.feature(TEXT_DOCUMENT_DOCUMENT_SYMBOL)
def document_symbol(server: GalaxyToolsLanguageServer, params: DocumentSymbolParams) -> Optional[List[DocumentSymbol]]:
    """Returns a list of symbols defined in the document."""
    document = _get_valid_document(server, params.text_document.uri)
    if document:
        xml_document = _get_xml_document(document)
        return server.service.symbols_provider.get_document_symbols(xml_document)
    return None


@language_server.command(Commands.AUTO_CLOSE_TAGS)
def auto_close_tag(server: GalaxyToolsLanguageServer, parameters: CommandParameters) -> Optional[AutoCloseTagResult]:
    """Responds to a close tag request to close the currently opened node."""
    if server.configuration.completion.auto_close_tags and parameters:
        params = convert_to(parameters[0], TextDocumentPositionParams)
        document = _get_valid_document(server, params.text_document.uri)
        if document:
            xml_document = _get_xml_document(document)
            return server.service.get_auto_close_tag(xml_document, params.position)
    return None


@language_server.command(Commands.GENERATE_TESTS)
async def cmd_generate_test(
    server: GalaxyToolsLanguageServer, parameters: CommandParameters
) -> Optional[GeneratedSnippetResult]:
    """Generates some test snippets based on the inputs and outputs of the document."""
    params = convert_to(parameters[0], TextDocumentIdentifier)
    document = _get_valid_document(server, params.uri)
    if document:
        return server.service.generate_tests(document)
    return None


@language_server.command(Commands.GENERATE_COMMAND)
async def cmd_generate_command(
    server: GalaxyToolsLanguageServer, parameters: CommandParameters
) -> Optional[GeneratedSnippetResult]:
    """Generates a boilerplate Cheetah code snippet based on the inputs and outputs of the document."""
    params = convert_to(parameters[0], TextDocumentIdentifier)
    document = _get_valid_document(server, params.uri)
    if document:
        return server.service.generate_command(document)
    return None


@language_server.command(Commands.SORT_SINGLE_PARAM_ATTRS)
def sort_single_param_attrs_command(
    server: GalaxyToolsLanguageServer, parameters: CommandParameters
) -> Optional[ReplaceTextRangeResult]:
    """Sorts the attributes of the param element under the cursor."""
    params = convert_to(parameters[0], TextDocumentPositionParams)
    document = _get_valid_document(server, params.text_document.uri)
    if document:
        xml_document = _get_xml_document(document)
        return server.service.sort_single_param_attrs(xml_document, params.position)
    return None


@language_server.command(Commands.SORT_DOCUMENT_PARAMS_ATTRS)
def sort_document_params_attrs_command(
    server: GalaxyToolsLanguageServer, parameters: CommandParameters
) -> Optional[List[ReplaceTextRangeResult]]:
    """Sorts the attributes of all the param elements contained in the document."""
    params = convert_to(parameters[0], TextDocumentIdentifier)
    document = _get_valid_document(server, params.uri)
    if document:
        xml_document = _get_xml_document(document)
        return server.service.sort_document_param_attributes(xml_document)
    return None


@language_server.command(Commands.GENERATE_EXPANDED_DOCUMENT)
def generate_expanded_command(server: GalaxyToolsLanguageServer, parameters: CommandParameters) -> GeneratedExpandedDocument:
    """Generates a expanded version (with all macros replaced) of the tool document."""
    params = convert_to(parameters[0], TextDocumentIdentifier)
    document = server.workspace.get_document(params.uri)
    if document and DocumentValidator.is_tool_document(document):
        return server.service.macro_expander.generate_expanded_from(document.path)
    return GeneratedExpandedDocument(error_message=f"The document {document.filename} is not a valid Galaxy Tool wrapper.")


@language_server.command(Commands.DISCOVER_TESTS_IN_WORKSPACE)
def discover_tests_in_workspace_command(
    server: GalaxyToolsLanguageServer, parameters: CommandParameters
) -> List[TestSuiteInfoResult]:
    """Returns a list of test suites, one for each tool file in the workspace."""
    return server.service.test_discovery_service.discover_tests_in_workspace(server.workspace)


@language_server.command(Commands.DISCOVER_TESTS_IN_DOCUMENT)
def discover_tests_in_document_command(
    server: GalaxyToolsLanguageServer, parameters: CommandParameters
) -> Optional[TestSuiteInfoResult]:
    """Returns a test suite containing all tests for a particular XML tool document."""
    params = convert_to(parameters[0], TextDocumentIdentifier)
    document = _get_valid_document(server, params.uri)
    if document:
        xml_document = _get_xml_document(document)
        return server.service.test_discovery_service.discover_tests_in_document(xml_document)
    return None


def _validate(server: GalaxyToolsLanguageServer, params) -> None:
    """Validates the Galaxy tool and reports any problem found."""
    diagnostics: List[Diagnostic] = []
    document = _get_valid_document(server, params.text_document.uri)
    if document:
        xml_document = _get_xml_document(document)
        diagnostics = server.service.get_diagnostics(xml_document)
    server.publish_diagnostics(params.text_document.uri, diagnostics)


def _get_valid_document(server: GalaxyToolsLanguageServer, uri: str) -> Optional[Document]:
    document = server.workspace.get_document(uri)
    if _is_document_supported(document):
        return document
    return None


def _get_xml_document(document: Document) -> XmlDocument:
    """Parses the input Document and returns an XmlDocument."""
    xml_document = XmlDocumentParser().parse(document)
    return xml_document


def _is_document_supported(document: Document) -> bool:
    """Returns True if the given document is supported by the server."""
    if not document.uri.lower().endswith(".xml"):
        return False
    return DocumentValidator.has_valid_root(document)



"""Service in charge of formating XML files following
best practices.
"""

from typing import List

from lsprotocol.types import (
    DocumentFormattingParams,
    Position,
    Range,
    TextEdit,
)
from lxml import etree

DEFAULT_INDENTATION = " " * 4


class GalaxyToolFormatService:
    """Galaxy tool format service.

    This service manages XML file formatting
    following best practices for Galaxy tools.
    """

    def format(self, content: str, params: DocumentFormattingParams) -> List[TextEdit]:
        """Given the document contents returns the list of TextEdits
        needed to properly layout the document.
        """
        formatted_result = self.format_content(content, params.options.tab_size)

        lines = content.count("\n")
        start = Position(line=0, character=0)
        end = Position(line=lines + 1, character=0)
        return [TextEdit(range=Range(start=start, end=end), new_text=formatted_result)]

    def format_content(self, content: str, tabSize: int = 4) -> str:
        """Formats the given XML content."""
        try:
            parser = etree.XMLParser(strip_cdata=False)
            xml = etree.fromstring(content, parser=parser)
            spaces = " " * tabSize
            etree.indent(xml, space=spaces)
            result = etree.tostring(xml, pretty_print=True, encoding=str)
            return result
        except etree.XMLSyntaxError:
            return content  # Do not auto-format if there are syntax errors



"""Module for converting XSD schema to tree structure
to simplify access to definitions.
"""

from typing import (
    cast,
    Dict,
    List,
    Optional,
)

from lxml import etree

from .constants import (
    MAX_RECURSION_DEPTH,
    XS_ALL,
    XS_ATTRIBUTE,
    XS_ATTRIBUTE_GROUP,
    XS_CHOICE,
    XS_COMPLEX_CONTENT,
    XS_COMPLEX_TYPE,
    XS_ELEMENT,
    XS_EXTENSION,
    XS_GROUP,
    XS_SEQUENCE,
    XS_SIMPLE_CONTENT,
    XS_SIMPLE_TYPE,
)
from .types import (
    XsdAttribute,
    XsdNode,
    XsdTree,
)


class GalaxyToolXsdParser:
    """Parser for converting the XSD schema into a tree structure
    that it is easy to use to retrieve information about the schema.

    Note: the conversion is focused in supporting the definitions
    used in the Galaxy tool's XSD and not all the possible definitions
    supported by the XSD specification.
    """

    def __init__(self, xsd_root: etree._Element):
        """Initializes the parser with the given etree root element.

        Args:
            xsd_root (etree._Element): The root element of the XSD.
        """
        self._root: etree._Element = xsd_root
        self._tree: Optional[XsdTree] = None
        self._named_type_map: Dict[str, etree._Element] = {}
        self._named_group_map: Dict[str, etree._Element] = {}

    def get_tree(self) -> XsdTree:
        """Builds the tree structure from the root etree._Element if
        it does not exists, or returns it if exists.

        Returns:
            XsdTree: The tree structure with all the relevant
            information from the XSD schema.
        """
        if self._tree is None:
            self._register_named_types()
            self._register_named_groups()
            self._build_tree_recursive(self._root, self._tree)
            if self._tree is None:
                self._tree = self._build_empty_tree()
        return self._tree

    def _build_empty_tree(self) -> XsdTree:
        return XsdTree(XsdNode(""))

    def _register_named_types(self) -> None:
        """Finds all (named) complex and simple type definitions in the
        XSD schema and register them in a dictionary.
        """
        for element in self._root.findall(XS_COMPLEX_TYPE):
            name = element.get("name")
            if name:
                self._named_type_map[name] = element

        for element in self._root.findall(XS_SIMPLE_TYPE):
            name = element.get("name")
            if name:
                self._named_type_map[name] = element

    def _register_named_groups(self) -> None:
        """Finds all (named) groups and attribute group definitions in the
        XSD schema and register them in a dictionary.
        """
        for element in self._root.findall(XS_GROUP):
            name = element.get("name")
            if name:
                self._named_group_map[name] = element

        for element in self._root.findall(XS_ATTRIBUTE_GROUP):
            name = element.get("name")
            if name:
                self._named_group_map[name] = element

    def _build_tree_recursive(self, parent_element: etree._Element, parent_node: Optional[XsdNode], depth: int = 0) -> None:
        if depth > MAX_RECURSION_DEPTH:
            return None  # Stop recursion

        for element in parent_element:
            tag = element.tag
            element_name = cast(Optional[str], element.attrib.get("name"))
            element_type_name = self._get_attribute_value_by_name(element, "type")
            if tag == XS_ELEMENT and element_name is not None:
                node = XsdNode(element_name, element, parent_node)
                if not parent_node:  # Is root element
                    self._tree = XsdTree(node)
                node.xsd_type = self._named_type_map.get(element_type_name)
                self._apply_named_type_to_node(element_type_name, node, depth + 1)
                # minOccurs defaults to 1
                min_occurs = cast(str, element.attrib.get("minOccurs", "1"))
                node.min_occurs = int(min_occurs)
                max_occurs = cast(str, element.attrib.get("maxOccurs", "unbounded"))
                node.max_occurs = int(max_occurs) if max_occurs != "unbounded" else -1
                self._build_tree_recursive(element, node, depth + 1)
            elif tag == XS_COMPLEX_TYPE:
                if not element_name and parent_node:
                    # The type is anonymous and is declared inside the element
                    # so we can directly apply it to the node
                    self._apply_complex_type_to_node(element, parent_node, depth + 1)
            elif tag == XS_GROUP:
                element_ref = cast(Optional[str], element.attrib.get("ref"))
                if element_ref and parent_node:
                    self._apply_group_to_node(element_ref, parent_node, depth + 1)

    def _apply_named_type_to_node(self, type_name: str, node: XsdNode, depth: int = 0) -> None:
        element_type = self._named_type_map.get(type_name)
        if element_type is not None and element_type.tag == XS_COMPLEX_TYPE:
            self._apply_complex_type_to_node(element_type, node, depth + 1)

    def _apply_complex_type_to_node(self, complex_type: etree._Element, node: XsdNode, depth: int = 0) -> None:
        for child_element in complex_type:
            tag = child_element.tag
            if tag in [XS_ALL, XS_SEQUENCE, XS_CHOICE]:
                self._build_tree_recursive(child_element, node, depth + 1)
            elif tag == XS_ATTRIBUTE:
                self._add_attribute_to_node(child_element, node)
            elif tag == XS_ATTRIBUTE_GROUP:
                element_ref = cast(Optional[str], child_element.attrib.get("ref"))
                if element_ref:
                    self._apply_attribute_group_to_node(element_ref, node)
            elif tag == XS_SIMPLE_CONTENT:
                self._apply_simple_content_to_node(child_element, node)
            elif tag == XS_COMPLEX_CONTENT:
                self._apply_complex_content_to_node(child_element, node, depth)

    def _apply_attribute_group_to_node(self, group_name: str, node: XsdNode) -> None:
        group = self._named_group_map.get(group_name)
        if group is not None:
            for child_element in group:
                self._add_attribute_to_node(child_element, node)

    def _apply_group_to_node(
        self,
        group_name: str,
        node: XsdNode,
        depth: int,
    ) -> None:
        group = self._named_group_map.get(group_name)
        if group is not None:
            for child_element in group:
                self._build_tree_recursive(child_element, node, depth + 1)

    def _apply_simple_content_to_node(self, simple_content: etree._Element, node: XsdNode) -> None:
        for child_elem in simple_content:
            if child_elem.tag == XS_EXTENSION:
                attributes = child_elem.findall(XS_ATTRIBUTE)
                for attr_elem in attributes:
                    self._add_attribute_to_node(attr_elem, node)

    def _apply_complex_content_to_node(self, complex_content: etree._Element, node: XsdNode, depth: int) -> None:
        for child_elem in complex_content:
            if child_elem.tag == XS_EXTENSION:
                attr_base = self._get_attribute_value_by_name(child_elem, "base")
                if attr_base:
                    self._apply_named_type_to_node(attr_base, node, depth)
                self._apply_complex_type_to_node(child_elem, node, depth)

    def _add_attribute_to_node(self, attribute_element: etree._Element, node: XsdNode) -> None:
        attr_name = self._get_attribute_value_by_name(attribute_element, "name")
        attr_type = self._get_attribute_value_by_name(attribute_element, "type")
        attr_use = self._get_attribute_value_by_name(attribute_element, "use")
        attr = XsdAttribute(attr_name, attribute_element, attr_type, attr_use == "required")
        attr.enumeration = self._get_enumeration_restrictions_from_type(attr_type)
        node.attributes[attr_name] = attr

    def _get_enumeration_restrictions_from_type(self, type_name: str) -> List[str]:
        simple_type = self._named_type_map.get(type_name)
        if simple_type is not None and simple_type.tag == XS_SIMPLE_TYPE:
            enumeration_values = simple_type.xpath(".//xs:enumeration/@value", namespaces=simple_type.nsmap)  # type: ignore
            return cast(List[str], enumeration_values)
        return []

    def _get_attribute_value_by_name(self, element: etree._Element, name: str) -> str:
        return cast(str, element.attrib.get(name, ""))



"""Constants definitions related to tool wrappers."""

TOOL = "tool"
INPUTS = "inputs"
PARAM = "param"
CONDITIONAL = "conditional"
REPEAT = "repeat"
SECTION = "section"
NAME = "name"
TYPE = "type"
SELECT = "select"
OPTION = "option"
VALUE = "value"
WHEN = "when"
TEST = "test"
TESTS = "tests"
TEXT = "text"
MIN = "min"
MAX = "max"
INTEGER = "integer"
FLOAT = "float"
BOOLEAN = "boolean"
TRUEVALUE = "truevalue"
FALSEVALUE = "falsevalue"
CHECKED = "checked"
ARGUMENT = "argument"
BOOLEAN_OPTIONS = ["true", "false", "True", "False", "yes", "no", "1", "0"]
EXPECT_NUM_OUTPUTS = "expect_num_outputs"
OUTPUTS = "outputs"
DATA = "data"
COLLECTION = "collection"
OUTPUT = "output"
OUTPUT_COLLECTION = "output_collection"
ASSERT_CONTENTS = "assert_contents"
HAS_TEXT = "has_text"
HAS_LINE = "has_line"
HAS_LINE_MATCHING = "has_line_matching"
HAS_N_COLUMNS = "has_n_columns"
HAS_SIZE = "has_size"
LINE = "line"
EXPRESSION = "expression"
N = "n"
DELTA = "delta"
ELEMENT = "element"
DASH = "-"
UNDERSCORE = "_"
DESCRIPTION = "description"
XREF = "xref"
COMMAND = "command"
ENV_VARIABLES = "environment_variables"
CONFIGFILES = "configfiles"
OPTIONAL = "optional"
LABEL = "label"
HELP = "help"
FORMAT = "format"
MACROS = "macros"
IMPORT = "import"
TOKEN = "token"
MACRO = "macro"
XML = "xml"



from typing import (
    Any,
    cast,
    Dict,
    List,
    Optional,
)

from anytree.search import findall
from galaxy.util import xml_macros
from lsprotocol.types import (
    Position,
    Range,
)
from lxml import etree
from pygls.workspace import Document

from galaxyls.constants import DEFAULT_DOCUMENT_RANGE
from .nodes import (
    XmlContainerNode,
    XmlElement,
    XmlSyntaxNode,
)
from .types import (
    DocumentType,
    NodeType,
)
from .utils import (
    convert_document_offset_to_position,
    convert_document_offsets_to_range,
)

MACRO_RELATED_TAGS = ["import", "token", "macro", "xml", "expand"]


class XmlDocument(XmlSyntaxNode):
    """Represents a parsed XML document.

    This is the root of the XML syntax tree.
    """

    DEFAULT_RANGE = DEFAULT_DOCUMENT_RANGE

    def __init__(self, document: Document):
        super().__init__()
        self.document: Document = document
        self.supported_document_types: Dict[str, DocumentType] = {
            "tool": DocumentType.TOOL,
            "macros": DocumentType.MACROS,
        }
        self._xml_tree: Optional[etree._ElementTree] = None
        self._xml_tree_expanded: Optional[etree._ElementTree] = None

    @property
    def node_type(self) -> NodeType:
        """The type of this node."""
        return NodeType.DOCUMENT

    @property
    def is_empty(self) -> bool:
        """True if the document has no root node."""
        return not self.root

    @property
    def root(self) -> Optional[XmlElement]:
        """The root element of the document.

        Normally this would be tool, macros or any other supported
        kind of element."""
        if len(self.children) == 0:
            return None
        try:
            return next(child for child in self.children if type(child) == XmlElement)
        except StopIteration:
            return None

    @property
    def uses_macros(self) -> bool:
        """Indicates if this XML document contains any <expand> element.

        Returns:
            bool: True if the tool contains at least one <expand> elements.
        """
        try:
            found = findall(self.root, filter_=lambda node: node.name in MACRO_RELATED_TAGS, mincount=1)
            return len(found) > 0
        except BaseException:
            return False

    @property
    def document_type(self) -> DocumentType:
        """The type of this document (if it is supported) or UNKNOWN."""
        if self.root and self.root.name:
            return self.supported_document_types.get(self.root.name, DocumentType.UNKNOWN)
        return DocumentType.UNKNOWN

    @property
    def is_unknown(self) -> bool:
        """Indicates if the document is of unknown type."""
        return self.document_type == DocumentType.UNKNOWN

    @property
    def is_macros_file(self) -> bool:
        """Indicates if the document is a macro definition file."""
        return self.document_type == DocumentType.MACROS

    @property
    def is_tool_file(self) -> bool:
        """Indicates if the document is a tool definition file."""
        return self.document_type == DocumentType.TOOL

    @property
    def xml_tree(self) -> Optional[etree._ElementTree]:
        """Internal XML tree structure."""
        if self._xml_tree is None:
            try:
                tree = etree.parse(self.document.path)
                self._xml_tree = tree
            except etree.XMLSyntaxError:
                pass  # Invalid XML document
        return self._xml_tree

    @property
    def xml_has_syntax_errors(self) -> bool:
        return self.xml_tree is None

    @property
    def xml_tree_expanded(self) -> Optional[etree._ElementTree]:
        """Internal XML tree structure after expanding macros.

        If there are no macro definitions, it returns the same as `xml_tree` property."""
        if self._xml_tree_expanded is None:
            if self.uses_macros:
                try:
                    expanded_tool_tree, _ = xml_macros.load_with_references(self.document.path)
                    self._xml_tree_expanded = expanded_tool_tree
                except etree.XMLSyntaxError:
                    pass  # Invalid XML document
                except BaseException:
                    pass  # TODO: Errors expanding macros should be catch during validation
            if self._xml_tree_expanded is None:
                # Fallback to the non-expanded version if something failed
                self._xml_tree_expanded = self.xml_tree
        return self._xml_tree_expanded

    def get_node_at(self, offset: int) -> Optional[XmlSyntaxNode]:
        """Gets the syntax node a the given offset."""
        return self.root.find_node_at(offset) if self.root else None

    def get_content_range(self, element: XmlContainerNode) -> Optional[Range]:
        """Gets the Range positions for the given XML element's content in the document.

        Args:
            element (XmlContainerNode): The XML element to determine it's content range positions.

        Returns:
            Optional[Range]: The range positions for the content of the given XML element.
        """
        start_offset, end_offset = element.get_content_offsets()
        if start_offset < 0 or end_offset < 0:
            return None
        return convert_document_offsets_to_range(self.document, start_offset, end_offset)

    def get_element_name_range(self, element: XmlElement) -> Optional[Range]:
        """Gets the Range positions for the given XML element's name in the document.

        Args:
            element (XmlElement): The XML element to determine it's name range positions.

        Returns:
            Optional[Range]: The range positions for the name of the given XML element.
        """
        start_offset = element.name_start_offset
        end_offset = element.name_end_offset
        if start_offset < 0 or end_offset < 0:
            return None
        return convert_document_offsets_to_range(self.document, start_offset, end_offset)

    def get_full_range(self, node: XmlSyntaxNode) -> Optional[Range]:
        """Gets the Range positions for the given XML node in the document.

        Args:
            node (XmlSyntaxNode): The XML node to determine it's range positions.

        Returns:
            Optional[Range]: The range positions for the entire node.
        """
        if node.start < 0 or node.end < 0:
            return None
        return convert_document_offsets_to_range(self.document, node.start, node.end)

    def get_position_before(self, element: XmlElement) -> Position:
        """Return the position in the document before the given element.

        Args:
            element (XmlElement): The element used to find the position.

        Returns:
            Position: The position just before the element declaration.
        """
        return convert_document_offset_to_position(self.document, element.start)

    def get_position_after(self, element: XmlElement) -> Position:
        """Return the position in the document after the given element.

        Args:
            element (XmlElement): The element used to find the position.

        Returns:
            Position: The position just after the element declaration.
        """
        if element.is_self_closed:
            return convert_document_offset_to_position(self.document, element.end)
        return convert_document_offset_to_position(self.document, element.end_offset)

    def get_position_after_last_child(self, element: XmlElement) -> Position:
        """Return the position in the document after the last child of the given element.

        Args:
            element (XmlElement): The element used to find the position.

        Returns:
            Position: The position just after the last child element declaration.
        """
        if element.is_self_closed:
            return convert_document_offset_to_position(self.document, element.end)
        if element.elements:
            last = element.elements[-1]
            return self.get_position_after(last)
        return convert_document_offset_to_position(self.document, element.end_tag_open_offset)

    def get_position_before_first_child(self, element: XmlElement) -> Position:
        """Return the position in the document before the first child of the given element.

        Args:
            element (XmlElement): The element used to find the position.

        Returns:
            Position: The position just before the first child element declaration.
        """
        if element.is_self_closed:
            return convert_document_offset_to_position(self.document, element.end)
        return convert_document_offset_to_position(self.document, element.end_tag_open_offset)

    def find_all_elements_with_name(self, name: str) -> List[XmlElement]:
        """Returns a list with all the elements contained in the document matching the given name."""
        if self.root:
            found = findall(self.root, filter_=lambda node: isinstance(node, XmlElement) and node.name == name)
            return list(found)
        return []

    def get_text_between_offsets(self, start: int, end: int) -> str:
        """Gets the text content between the start and end offsets."""
        return self.document.source[start:end]

    def get_text_in_range(self, range: Range) -> str:
        """Gets the document text contained in the given range."""
        start = self.document.offset_at_position(range.start)
        end = self.document.offset_at_position(range.end)
        return self.get_text_between_offsets(start, end)

    def get_line_indentation(self, line_number: int) -> str:
        """Gets the string containing the number of spaces at the beginning of the given line in the document."""
        line_text = self.document.lines[line_number]
        indentation = line_text[: len(line_text) - len(line_text.lstrip())]
        return indentation

    def get_default_range(self) -> Range:
        """Gets the range of the root tag or the first character of the document if there is no root."""
        if self.root:
            return self.get_element_name_range(self.root) or self.DEFAULT_RANGE
        return self.DEFAULT_RANGE

    def get_tree_element_from_xpath(self, tree, xpath: Optional[str]) -> Optional[Any]:
        if xpath is None or tree is None:
            return None
        element: Any = tree.xpath(xpath)
        if element is not None:
            if isinstance(element, list):
                element = cast(list, element)
                if len(element) > 0:
                    return element[0]
        return None

    def get_element_from_xpath(self, xpath: Optional[str]) -> Optional[Any]:
        return self.get_tree_element_from_xpath(self.xml_tree, xpath)

    def get_internal_element_range_or_default(self, element: Optional[Any]) -> Range:
        if element is not None:
            line_number = element.sourceline - 1
            line_text = self.document.lines[line_number]
            if isinstance(element, etree._Comment):
                text = cast(str, element.text)
                start = line_text.index(f"{text}")
                end = start + len(text)
            else:
                # Prepend '<' for searching tag names
                start = line_text.index(f"<{element.tag}") + 1
                end = start + len(element.tag)
            return Range(
                start=Position(line=line_number, character=start),
                end=Position(line=line_number, character=end),
            )
        return self.get_default_range()

    def get_element_range_from_xpath_or_default(self, xpath: Optional[str]) -> Range:
        element = self.get_element_from_xpath(xpath)
        return self.get_internal_element_range_or_default(element)



""" Type definitions for XSD processing.
"""

from typing import (
    Any,
    cast,
    Dict,
    List,
    Optional,
)

from anytree import (
    findall,
    NodeMixin,
    RenderTree,
    Resolver,
    ResolverError,
)
from lsprotocol.types import (
    MarkupContent,
    MarkupKind,
)
from lxml import etree

from .constants import MSG_NO_DOCUMENTATION_AVAILABLE


class XsdBase:
    """Base class that encapsulates an element from the XSD schema.

    It contains common information and functionality shared by
    XML nodes and attributes.
    """

    def __init__(self, name: str, element: Optional[etree._Element]):
        super(XsdBase, self).__init__()
        self.name: str = name
        self.xsd_element: Optional[etree._Element] = element
        self.xsd_type: Optional[etree._Element] = None

    def __repr__(self) -> str:
        return self.name

    def get_doc(self, lang: str = "en") -> MarkupContent:
        """Gets the Markdown documentation associated with this element
        from the XSD schema.

        If there is no documentation in the schema for the element,
        a message indicating this will be returned instead.

        Args:
            lang (str, optional): The language code of the documentation
            to retrieve. Defaults to "en" (English).

        Returns:
            [str]: The documentation text or a message indicating
            there is no documentation.
        """
        doc = self._get_doc_text_of_element(self.xsd_element, lang) or self._get_doc_text_of_element(self.xsd_type, lang)
        if doc:
            return MarkupContent(kind=MarkupKind.Markdown, value=doc)
        return MarkupContent(kind=MarkupKind.Markdown, value=MSG_NO_DOCUMENTATION_AVAILABLE)

    def _get_doc_text_of_element(self, element: Optional[Any], lang: str = "en") -> str:
        try:
            if element is not None:
                doc_annotation = element.xpath(
                    "./xs:annotation/xs:documentation[@xml:lang=$lang]/text()",
                    namespaces=element.nsmap,
                    lang=lang,
                )
                return cast(List[str], doc_annotation)[0].strip()
        except BaseException:
            pass
        return ""


class XsdAttribute(XsdBase):
    """Represents an attribute in an XML tag.

    It contains information about the attribute extracted
    from the XSD schema.

    Args:
        XsdBase: Inherits base functionality from XsdBase.
    """

    def __init__(
        self,
        name: str,
        element: Optional[etree._Element],
        type_name: Optional[str] = None,
        is_required: bool = False,
    ):
        super(XsdAttribute, self).__init__(name, element)
        self.type_name: Optional[str] = type_name
        self.is_required: bool = is_required
        self.enumeration: List[str] = []


class XsdNode(XsdBase, NodeMixin):
    """Represents a particular XML tag.

    The node contains information extracted from the XSD schema
    about the XML tag, possible descendants, attributes, etc.

    Args:
        XsdBase: Inherits base functionality from XsdBase.
        NodeMixin: Inherits tree node functionality from NodeMixin.
    """

    def __init__(self, name: str, element: Optional[etree._Element] = None, parent: Optional[NodeMixin] = None):
        super(XsdNode, self).__init__(name, element)
        self.parent: Optional[NodeMixin] = parent
        self.attributes: Dict[str, XsdAttribute] = {}
        self.min_occurs: int = 1  # required by default
        self.max_occurs: int = -1  # unbounded by default

    def render(self) -> str:
        """Gets an ascii representation of this node.

        Returns:
            str: An ascii representation of the node.
        """
        return str(RenderTree(self).by_attr(lambda node: f"[{node.name}] {' '.join(node.attributes)}"))


class XsdTree:
    """Represents a tree structure containing the important
    XSD information for all the elements and attributes.
    """

    def __init__(self, root: XsdNode):
        self.root: XsdNode = root
        self.node_resolver = Resolver("name")
        self.expand_element = self._build_expand_element()

    def find_node_by_stack(self, node_stack: List[str]) -> Optional[XsdNode]:
        """Finds the node definition in the tree that matches the given stack of tags.

        Args:
            node_stack (List[str]): The stack of tag names composing a tree branch.
            Like: ['root', 'node', 'subnode', 'leaf']

        Returns:
            Optional[XsdNode]: The node definition matching the leaf in the path or
            None if the node could not be found.
        """
        result_node = None
        if node_stack:
            try:
                path = self._get_path_from_stack(node_stack)
                if path.endswith(self.expand_element.name):
                    return self.expand_element
                result_node = self.node_resolver.get(self.root, path)
            except ResolverError as e:
                print(e)
        return result_node

    def render(self) -> str:
        """Gets an ascii representation of this tree.

        Returns:
            str: An ascii representation of the tree.
        """
        return self.root.render()

    def _get_path_from_stack(self, node_stack: List[str]) -> str:
        if node_stack[0] == self.root.name:
            node_stack[0] = "."
        return "/".join(node_stack)

    def _build_expand_element(self) -> XsdNode:
        """Creates a XsdNode representing the <expand> element and it's attributes.
        The <expand> element is a special element that is not defined in the XSD schema but allows to use macros in the tool.

        Returns:
            XsdNode: The node with information about the special <expand> element.
        """
        expand_node = XsdNode("expand", None)
        attr_name = "macro"
        attr = XsdAttribute(attr_name, None, type_name=None, is_required=True)
        expand_node.attributes[attr_name] = attr
        return expand_node

    def find_node_by_name(self, name: str) -> Optional[XsdNode]:
        if name == self.expand_element.name:
            return self.expand_element

        result = findall(
            self.root,
            filter_=lambda node: node.name == name,
        )
        if len(result) > 0:
            return cast(XsdNode, result[0])
        return None



from typing import (
    cast,
    List,
    Optional,
    Tuple,
    Union,
)

from lsprotocol.types import (
    Position,
    Range,
)
from lxml import etree

from galaxyls.services.tools.constants import (
    ARGUMENT,
    ASSERT_CONTENTS,
    BOOLEAN,
    BOOLEAN_OPTIONS,
    CHECKED,
    COLLECTION,
    CONDITIONAL,
    DATA,
    DELTA,
    ELEMENT,
    EXPECT_NUM_OUTPUTS,
    EXPRESSION,
    HAS_LINE,
    HAS_LINE_MATCHING,
    HAS_N_COLUMNS,
    HAS_SIZE,
    HAS_TEXT,
    INPUTS,
    LINE,
    N,
    NAME,
    OPTION,
    OUTPUT,
    OUTPUT_COLLECTION,
    OUTPUTS,
    PARAM,
    REPEAT,
    SECTION,
    SELECT,
    TEST,
    TESTS,
    TEXT,
    TOOL,
    TYPE,
    VALUE,
)
from galaxyls.services.tools.document import GalaxyToolXmlDocument
from galaxyls.services.tools.generators.snippets import SnippetGenerator
from galaxyls.services.tools.inputs import (
    ConditionalInputNode,
    InputNode,
    RepeatInputNode,
    SectionInputNode,
)
from galaxyls.services.xml.nodes import XmlElement

AUTO_GEN_TEST_COMMENT = "TODO: auto-generated test case. Please fill in the required values"
BOOLEAN_CONDITIONAL_NOT_RECOMMENDED_COMMENT = (
    "Warning: the use of boolean as a conditional parameter is not recommended. Please consider using a select instead."
)


class GalaxyToolTestSnippetGenerator(SnippetGenerator):
    """This class tries to generate the XML code for a test case using the information
    already defined in the inputs and outputs of the tool XML wrapper.
    """

    def __init__(self, tool_document: GalaxyToolXmlDocument, tabSize: int = 4) -> None:
        super().__init__(tool_document, tabSize)

    def _build_snippet(self) -> Tuple[str, bool]:
        """This function tries to generate a code snippet in TextMate format with all the tests cases extracted
        from the inputs and outputs of the tool.

        Returns:
            Tuple[str, bool]: The code snippet in TextMate format or an error message if the generation failed.
            The second value of the tuple indicates if there was an error.
        """
        try:
            input_tree = self.expanded_document.analyze_inputs()
            outputs = self.expanded_document.get_outputs()
            result_snippet = "\n".join(
                (self._generate_test_case_snippet(input_node, outputs) for input_node in input_tree.leaves)
            )
            tests_section = self.tool_document.find_element(TESTS)
            if tests_section and not tests_section.is_self_closed:
                return (result_snippet, False)
            return (f"\n<{TESTS}>\n{result_snippet}\n</{TESTS}>", False)
        except BaseException as ex:
            return (f"Automatic Test Case generation failed with reason: {ex}", True)

    def _find_snippet_insert_position(self) -> Union[Position, Range]:
        """Returns the position inside the document where new test cases
        can be inserted.

        If the <tests> section does not exists in the file, the best aproximate
        position where the tests should be inserted is returned (acording to the IUC
        best practices tag order).

        Returns:
            Position: The position where the test cases can be inserted in the document.
        """
        tool = self.tool_document
        section = tool.find_element(TESTS)
        if section:
            content_range = tool.get_content_range(section)
            if content_range:
                return content_range.end
            else:  # is self closed <tests/>
                start = tool.get_position_before(section)
                end = tool.get_position_after(section)
                return Range(start=start, end=end)
        else:
            section = tool.find_element(OUTPUTS)
            if section:
                return tool.get_position_after(section)
            section = tool.find_element(INPUTS)
            if section:
                return tool.get_position_after(section)
            section = tool.find_element(TOOL)
            if section:
                content_range = tool.get_content_range(section)
                if content_range:
                    return content_range.end
            return Position(line=0, character=0)

    def _generate_test_case_snippet(self, input_node: InputNode, outputs: List[XmlElement]) -> str:
        """Generates the code snippet for a single <test> element given an InputNode and the list of outputs
        from the tool document.

        Args:
            input_node (InputNode): The InputNode that is one of the leaves of the input tree.
            outputs (List[XmlElement]): The list of XML elements representings the outputs of the tool.
            spaces (str, optional): The str with the number of spaces for an indent level. Defaults to "  ".

        Returns:
            str: The resulting code snippet in TextMate format.
        """
        test_element = self._create_test_element()
        self._add_inputs_to_test_element(input_node, test_element)
        self._add_outputs_to_test_element(outputs, test_element)
        etree.indent(test_element, space=self.indent_spaces)
        snippet = etree.tostring(test_element, pretty_print=True, encoding=str)
        return snippet

    def _create_test_element(self) -> etree._Element:
        """Returns a XML element representing a <test> tag with the basic information.

        Returns:
            etree._Element: The <test> XML element.
        """
        node = etree.Element(TEST)
        node.append(etree.Comment(AUTO_GEN_TEST_COMMENT))
        node.attrib[EXPECT_NUM_OUTPUTS] = self._get_next_tabstop()
        return node

    def _add_inputs_to_test_element(self, input_node: InputNode, parent: etree._Element) -> None:
        """Populates the given 'parent' XML element with the tests inputs of one of the input branches.

        Given an InputNode (usually a leaf node from the input tree) the path of nodes to this InputNode
        is used to create all the input XML elements that form this conditional branch of inputs.

        Args:
            input_node (InputNode): A leaf node of the input tree.
            parent (etree._Element): The parent XML element that will hold the input XML elements of this
            input branch. This is usually a <test> element.
        """
        current_parent = parent
        input_path = list(input_node.path)
        # 'input_path' contains the input nodes composing this conditional branch of inputs
        for node in input_path:
            node = cast(InputNode, node)
            if type(node) is ConditionalInputNode:
                conditional_element = self._build_conditional_test_element(node)
                current_parent.append(conditional_element)
                current_parent = conditional_element
            else:
                self._build_test_tree(node, current_parent)

    def _add_outputs_to_test_element(self, outputs: List[XmlElement], parent: etree._Element) -> None:
        """Converts the list of XML outputs to output test elements and appends them to the 'parent' element.

        Args:
            outputs (List[XmlElement]): The lists of XML outputs definitions of the tool.
            parent (etree._Element): The XML element that will hold the resulting test outputs. Usually the <test> tag.
        """
        for output in outputs:
            if output.name == DATA:
                self._add_output_to_test(output, parent)
            elif output.name == COLLECTION:
                self._add_output_collection_to_test(output, parent)

    def _build_param_test_element(self, input_param: XmlElement, value: Optional[str] = None) -> etree._Element:
        """Builds a <param> XML element to be used in a <test> code snippet from a given input <param> XML element.

        Args:
            input_param (XmlElement): The XML element of a <param> in the tool <inputs> section.
            value (Optional[str], optional): The optional value that will be assigned to the <param>. Defaults to None.

        Returns:
            etree._Element: The XML element representing a <param> in the <test> code snippet.
        """
        param = etree.Element(PARAM)
        argument_attr = input_param.get_attribute_value(ARGUMENT)
        name_attr = input_param.get_attribute_value(NAME)
        if not name_attr and argument_attr:
            name_attr = self._extract_name_from_argument(argument_attr)
        if name_attr:
            param.attrib[NAME] = name_attr
        default_value = input_param.get_attribute_value(VALUE)
        if value:
            param.attrib[VALUE] = value
        elif default_value:
            param.attrib[VALUE] = self._get_next_tabstop_with_placeholder(default_value)
        else:
            type_attr = input_param.get_attribute_value(TYPE)
            if type_attr:
                if type_attr == BOOLEAN:
                    default_value = input_param.get_attribute_value(CHECKED)
                    param.attrib[VALUE] = self._get_next_tabstop_with_options(BOOLEAN_OPTIONS, default_value)
                elif type_attr == SELECT or type_attr == TEXT:
                    try:
                        options = self._get_options_from_param(input_param)
                        param.attrib[VALUE] = self._get_next_tabstop_with_options(options, default_value)
                    except BaseException:
                        param.attrib[VALUE] = self._get_next_tabstop()
                else:
                    param.attrib[VALUE] = self._get_next_tabstop()
        return param

    def _build_conditional_test_element(self, input_conditional: ConditionalInputNode) -> etree._Element:
        """Builds a <conditional> XML element to be used in a <test> code snippet from a
        given input <conditional> XML element."""
        conditional = etree.Element(CONDITIONAL)
        conditional.attrib[NAME] = input_conditional.name
        if input_conditional.option_param:
            # add the option param
            param_element = self._build_param_test_element(input_conditional.option_param, input_conditional.option)
            if input_conditional.option_param.get_attribute_value(TYPE) == BOOLEAN:
                conditional.append(etree.Comment(BOOLEAN_CONDITIONAL_NOT_RECOMMENDED_COMMENT))
            conditional.append(param_element)
        # add the rest of params in the corresponding when element
        self._build_test_tree(input_conditional, conditional)
        return conditional

    def _build_min_repeat_test_elements(self, input_repeat: RepeatInputNode) -> List[etree._Element]:
        """Builds a <repeat> XML element to be used in a <test> code snippet from a given input <repeat> XML element."""
        repeats: List[etree._Element] = []
        for _ in range(input_repeat.min):
            repeat_node = etree.Element(REPEAT)
            repeat_node.attrib[NAME] = input_repeat.name
            self._build_test_tree(input_repeat, repeat_node)
            repeats.append(repeat_node)
        return repeats

    def _build_section_test_element(self, input_section: SectionInputNode) -> etree._Element:
        """Builds a <section> XML element to be used in a <test> code snippet from a given input <section> XML element."""
        section_node = etree.Element(SECTION)
        section_node.attrib[NAME] = input_section.name
        self._build_test_tree(input_section, section_node)
        return section_node

    def _build_test_tree(self, input: InputNode, parent: etree._Element) -> None:
        """Recursively adds to the 'parent' XML element all the input nodes from the given 'input'.

        Args:
            input (InputNode): The InputNode to extract the node information.
            parent (etree._Element): The XML element that will contain the elements of the generated tree.
        """
        for param in input.params:
            param_element = self._build_param_test_element(param)
            parent.append(param_element)
        for repeat in input.repeats:
            repeat_elements = self._build_min_repeat_test_elements(repeat)
            for rep in repeat_elements:
                parent.append(rep)
        for section in input.sections:
            section_element = self._build_section_test_element(section)
            parent.append(section_element)

    def _add_output_to_test(self, data: XmlElement, test_element: etree._Element) -> None:
        """Converts the given 'data' (<data>) XML element in an output XML element and adds it
        to the given <test> element.

        Args:
            output (XmlElement): The
            test_element (etree._Element): [description]
        """
        name = data.get_attribute_value(NAME)
        if name:
            output_element = etree.Element(OUTPUT)
            output_element.attrib[NAME] = name
            self._add_default_asserts_to_output(output_element)
            test_element.append(output_element)

    def _add_output_collection_to_test(self, output_collection: XmlElement, test_element: etree._Element) -> None:
        """Adds the 'output_collection' XML element to the 'test_element' with a default <element>.

        Args:
            output_collection (XmlElement): The <collection> XML element.
            test_element (etree._Element): The <test> XML element.
        """
        name = output_collection.get_attribute_value(NAME)
        if name:
            output_element = etree.Element(OUTPUT_COLLECTION)
            output_element.attrib[NAME] = name
            type_attr = output_collection.get_attribute_value(TYPE)
            if type_attr:
                output_element.attrib[TYPE] = type_attr
            element = etree.Element(ELEMENT)
            element.attrib[NAME] = self._get_next_tabstop()
            self._add_default_asserts_to_output(element)
            output_element.append(element)
            test_element.append(output_element)

    def _add_default_asserts_to_output(self, output: etree._Element) -> None:
        """Given a XML element (usually an output), this method adds the <assert_contents> tag
        with some examples of asserts.

        Args:
            output (etree._Element): The output XML element of the <test> that will contain
            the <assert_contents>.
        """
        assert_contents = etree.Element(ASSERT_CONTENTS)

        has_text = etree.Element(HAS_TEXT)
        has_text.attrib[TEXT] = self._get_next_tabstop()
        assert_contents.append(has_text)

        has_line = etree.Element(HAS_LINE)
        has_line.attrib[LINE] = self._get_next_tabstop()
        assert_contents.append(has_line)

        has_line_matching = etree.Element(HAS_LINE_MATCHING)
        has_line_matching.attrib[EXPRESSION] = self._get_next_tabstop()
        assert_contents.append(has_line_matching)

        has_n_columns = etree.Element(HAS_N_COLUMNS)
        has_n_columns.attrib[N] = self._get_next_tabstop()
        assert_contents.append(has_n_columns)

        has_size = etree.Element(HAS_SIZE)
        has_size.attrib[VALUE] = self._get_next_tabstop()
        has_size.attrib[DELTA] = self._get_next_tabstop()
        assert_contents.append(has_size)

        output.append(assert_contents)

    def _get_options_from_param(self, param: XmlElement) -> List[str]:
        """Gets the list of children elements of type <option> from the given 'param' XML element.

        Args:
            param (XmlElement): The <param> XML element.

        Returns:
            List[str]: The list of <option> in this 'param' or an empty list.
        """
        option_elements = param.get_children_with_name(OPTION)
        options = [o.get_attribute_value(VALUE) for o in option_elements]
        return list(filter(None, options))



from enum import Enum
from typing import Optional

import attrs


class CompletionMode(str, Enum):
    AUTO = "auto"
    INVOKE = "invoke"
    DISABLED = "disabled"


@attrs.define
class CompletionConfig:
    """Auto-completion feature configuration."""

    mode: CompletionMode = attrs.field(default=CompletionMode.AUTO)
    auto_close_tags: bool = attrs.field(default=True)


@attrs.define
class ServerConfig:
    """Language Server specific configuration."""

    silent_install: bool = attrs.field(default=False)


@attrs.define
class PlanemoTestingConfig:
    """Planemo testing configuration."""

    enabled: bool = attrs.field(default=True)
    auto_test_discover_on_save_enabled: bool = attrs.field(default=True)
    extra_params: str = attrs.field(default="")


@attrs.define
class PlanemoConfig:
    """Planemo integration configuration."""

    enabled: bool = attrs.field(default=False)
    env_path: str = attrs.field(default="planemo")
    galaxy_root: Optional[str] = attrs.field(default=None)
    get_cwd: Optional[str] = attrs.field(default=None)
    testing: PlanemoTestingConfig = attrs.field(default=PlanemoTestingConfig())


@attrs.define
class GalaxyToolsConfiguration:
    """Galaxy Language Server general configuration."""

    server: ServerConfig = attrs.field(default=ServerConfig())
    completion: CompletionConfig = attrs.field(default=CompletionConfig())
    planemo: PlanemoConfig = attrs.field(default=PlanemoConfig())
